use strict;

package GeneMethods;

use Gene;
use GenomicsUtils;
use lib "../../stat";
use lib "./stat";
use lib "../stat"; # a hack for when this is included from Genomics/seq or such
use Vector;

my %locuss;
sub readFile($) # XXX why is %locuss public!
{
    my $file = $_[0];
    open( F, $file );
    while( <F> ) {
	chomp;
	my ($scaffoldId, $locusId, $string) = ($_ =~m/^(\d+)\t(\d+)\t(.*)/os );
	my @list = split //, $string;
	$locuss{$locusId} = \@list;
    }
	
    close F;
}

sub mutualInfoPair( $ $ ) 
{
    my ($g1, $g2) = @_;
    my @list1 = @{ $locuss{$g1->locusId()} };
    my @list2 = @{ $locuss{$g2->locusId()} };
    
    my $len = @list1;
    my %count1=( 0=>0, 1=>0 );
    my %count2 =( 0=>0, 1=>0 );
    my %mutial = ( 00=>0, 01=>0, 10=>0, 11=>0 );
    for( my $i=0; $i < $len; ++$i ) {
	++$count1{ $list1[$i] };
	++$count2{ $list2[$i] };
	++$mutial{ $list1[$i].$list2[$i] };
    }

    my $nMismatches = $mutial{01} + $mutial{10};
    my @p1 = ( $count1{0}/$len, $count1{1}/$len );
    my @p2 = ( $count2{0}/$len, $count2{1}/$len );
    my @pm = ( $mutial{00}/$len, $mutial{01}/$len, $mutial{10}/$len, $mutial{11}/$len );

    # Calculate entropy
    my $H = 0;
    foreach my $i (0,1) { $H -= $p1[$i]==0 ? 0 : $p1[$i]*log( $p1[$i] ); }
    foreach my $i (0,1) { $H -= $p2[$i]==0 ? 0 : $p2[$i]*log( $p2[$i] ); }
    foreach my $i (0..3) {$H += $pm[$i]==0 ? 0 : $pm[$i]*log( $pm[$i] ); }
    
#    $H = floor( $H * 100000 + 0.5) / 100000;
    return ( $H, $nMismatches );
}

sub gcContent($)
{
    my $s = $_[0];
    return $s =~ tr/GCgc/XXxx/;
}

sub gcDelta( $ $ ) {
    my ($gene1, $gene2) = @_;
    my $seq1 = $gene1->dna();
    my $seq2 = $gene2->dna();
    my $gc1 = gcContent( $seq1 ) / length(  $seq1 );
    my $gc2 = gcContent( $seq2 ) / length(  $seq2 );

    my $delta = abs( $gc1 - $gc2 );

    return $delta;
}

# Determines the length of the intergenic region, taking circularity and strandedness
# into account and ignoring the order of the arguments.
# Assumes genes don't cross the origin
# Returns:
#   undef if the arguments are not on the same scaffold, or else
#   (separation,code), where separation is in bp and code is "Convergent", "Divergent", or "Same"
#   Separation is negative or 0 if they overlap (0 means they share 1 base)
#   WARNING: most papers on operon prediction use -1 instead of 0 to describe LocusS that overlap by 1 base
#
#   If two genes of differing orientations overlap, this labels it with the
#   code that leads to the least overlap. This is intuitively obvious for modest overlaps, but
#   seems arbitrary for extensive overlaps or if one gene is contained within the other.
#
sub geneSeparation( $ $ ) {
    my ($gene1, $gene2) = @_;
    return undef if ($gene1->scaffoldId() ne $gene2->scaffoldId());
    my $strand1 = $gene1->strand();
    my $strand2 = $gene2->strand();
    my $scaffold = $gene1->scaffold();

    my $overlap = $gene1->end() >= $gene2->begin() && $gene2->end() >= $gene1->begin();
    my ($separation, $code);
    if ($strand1 eq $strand2) {
	$code = "Same";
	$separation = Vector::min( abs($scaffold->distance($gene1->stop(), $gene2->start())),
				   abs($scaffold->distance($gene2->stop(), $gene1->start())) );
    } else {
	my $dStart = abs( $scaffold->distance($gene1->start(), $gene2->start()) );
	my $dStop = abs( $scaffold->distance($gene1->stop(), $gene2->stop()) );
	$code = $dStop < $dStart ? "Convergent" : "Divergent";
	$separation = $dStop < $dStart? $dStop : $dStart;
    }
    return ($separation * ($overlap ? -1 : 1), $code);
}

# Given a list of attributes of genes, finds the matching locus ids
#
# Starts with the list of hashes with attributes and a list of gene objects
# Currently, the only handled attributes are "synonym" and "VIMSS"
# It would be nice to add sequence and position-based matching...
#
# The third argument is a list of parameters for each attribute.
# Currently handled are:
# uniq: default 1 (true) -- fail on genes which do not match uniquely in both directions
# synonymType: default -1 (any) -- what synonym type code to allow (0=4-letter name, 1=org-specific, 2=gi)
# synonymCase: default 0 (false) -- is synonym matching case sensitive?
#
# Returns a list matching the original, with each hash containing the original attributes, plus
# locusId -> list of locusids
# success -> 1 or 0 (depending on uniq)
# nonuniq -> list of handle-sets which matched the same locuss, including self
# (non-uniquely means this handle found a gene which matched to a different handle from a different input row)

sub geneMatcher($$$) {
    my ($qRef,$genesRef,$paramsRef) = @_;

    my $uniq = defined $paramsRef->{"uniq"} ? $paramsRef->{"uniq"} : 1;
    my $synonymType = defined $paramsRef->{"synonymType"} ? $paramsRef->{"synonymType"} : -1;
    my $synonymCase = defined $paramsRef->{"synonymCase"} ? $paramsRef->{"synonymCase"} : 0;

    my %lookup = (); # attribute handle -> list of matching VIMSS id
    my %handleSets = (); # locusId -> matching handle-sets (sorted & joined by :::) -> 1

    foreach my $gene (@{ $genesRef }) {
	my $vimssHandle = "VIMSS" . $gene->locusId();
	&addToHashList(\%lookup, $vimssHandle, $gene);

	my $synonyms = $gene->synonym();
	foreach my $type (keys %{ $synonyms }) {
	    my $syn = $synonyms->{$type};
	    if ($synonymType == -1 || $synonymType == $type) {
		my $synHandle = $synonymCase ? "SYN" . $syn : "SYN" . lc($syn);
		&addToHashList(\%lookup, $synHandle, $gene->locusId());
	    }
	}
    }
    foreach my $list (values %lookup) {
	foreach my $locusId (@{ $list }) {
	    $handleSets{$locusId} = {};
	}
    }

    my @matches = (); # original hash plus "locusIds" -> list of locusIds, etc.

    foreach my $query (@{ $qRef }) {
	my @handles = ();
	if (exists $query->{"VIMSS"}) {
	    push @handles, "VIMSS" . $query->{"VIMSS"};
	}
	if (exists $query->{"synonym"}) {
	    push @handles, "SYN" . ($synonymCase ? $query->{"synonym"} : lc($query->{"synonym"}));
	}

	my $handleSet = join(":::", sort(@handles));

	my %locusIds = (); # locusId -> 1 if used
	foreach my $handle (@handles) {
	    if (exists $lookup{$handle}) {
		map { $locusIds{$_} = 1, $handleSets{$_}->{$handleSet} = 1 } @{ $lookup{$handle} };
	    }
	}
	my $match = {};
	map { $match->{$_} = $query->{$_} } (keys %{ $query });
	$match->{"locusIds"} = [];
	push @{ $match->{"locusIds"} }, (sort {$a <=> $b} keys %locusIds);
	$match->{"handles"} = [];
	push @{ $match->{"handles"} }, @handles;
	push @matches, $match;
    }

    foreach my $match (@matches) {
	my @duped = ();
	foreach my $locusId ( @{ $match->{"locusIds"} } ) {
	    push @duped, keys %{ $handleSets{$locusId} };
	}
	my @nonuniq = Vector::uniq(sort(@duped));
	@nonuniq = () if scalar @nonuniq == 1;
	$match->{"nonuniq"} = \@nonuniq;
	$match->{"success"} = $uniq ? ( scalar @{ $match->{"locusIds"} } == 1 && scalar @{ $match->{"nonuniq"} } == 0)
	    : (scalar @{ $match->{"locusIds"} } >= 1);
    }
    return \@matches;
}

1;
