package Genome;
require Exporter;


use strict;
use GenomicsUtils;
use DBI;
use Gene;
use Scaffold;

our (@ISA, @EXPORT);
 
@ISA = qw(Exporter);	
@EXPORT = qw( getGenomeMap getGenomeList lineage);

{

sub new;
# Constructor for gene object. Returns Genome object.
# Currently only takes (requires) "taxonomyId => xxx" as input.

#--------------------------------------------------------------
#The following functions return basic info about genomes and are
#relatively self-explanatory.

sub name;
sub taxonomyId;
sub lineage;
sub scaffolds;
# Returns a list of scaffold objects.  Only genomic (not plasmid)
# scaffolds are returned.
sub nongenomicScaffolds;
# Returns a list of nongenomics scaffold objects.

sub genes; # Returns a list of genes objects on genomic scaffolds
sub nongenomicGenes; # Returns a list of genes objects on non-genomic scaffolds
#--------------------------------------------------------------

sub getGenomeMap;
sub getGenomeList;

# These functions return a set of genome objects corresponding to all
# the genomes in the DB, wither as a list or as a hash keyed on
# taxonomyId.

#--------------------------------------------------------------

sub new{
    my %params = @_;
    my $taxonomyId = $params{'taxonomyId'};
    my $self = {};
    bless $self;
    
    my @scaffoldIds = &queryList( qq{ SELECT scaffoldId FROM Scaffold WHERE taxonomyId=$taxonomyId AND isActive=1 } );

    my @scaffolds;
    foreach my $sId (@scaffoldIds){
	push @scaffolds, Scaffold::new( scaffoldId => $sId );
    }

    $self->{scaffolds_} = \@scaffolds;
    $self->{taxonomyId_} = $taxonomyId;
    if (defined $params{'name'}){
	$self->{name_} = $params{'name'};
    }else{
	$self->{name_} = &queryScalar( "SELECT name FROM Taxonomy WHERE taxonomyId=$taxonomyId" );
    }

    return $self;
}

sub name{
    my $self = shift;
    return $self->{name_};
}

sub taxonomyId{
    my $self = shift;
    return $self->{taxonomyId_};
}
#my %lineage = ();
sub lineage{
    my ($taxId) = @_;
    my $parentR = GenomicsUtils::query(qq{SELECT T.rank, t.name, T.parentId FROM TaxNode T, TaxName t WHERE t.taxonomyId=T.taxonomyId AND t.class="Scientific name" AND T.taxonomyId=$taxId});
    my ($rank, $name, $parentId) = @{$parentR->[0]};
    #$lineage{$parentRank} = $parentName;
    my @lineage = ($rank."=>".$name); 
    unshift @lineage, &lineage($parentId) unless $parentId==131567;  #taxonomy ID 131567: cellular organisms
    return @lineage;
}

sub scaffolds{
    my $self = shift;
    return grep {$_->isGenomic()} @{$self->{scaffolds_}};
}

sub nongenomicScaffolds{
    my $self = shift;
    return grep {! $_->isGenomic()} @{$self->{scaffolds_}};
}

# Passes along arguments, ultimately reaching Scaffold::fetchGenesListRef
sub genes{
    my $self = shift;
    return map {$_->genes(@_)} $self->scaffolds();
}

# Retrieve plasmid genes
sub nongenomicGenes{
    my $self = shift;
    return map {$_->genes(@_)} $self->nongenomicScaffolds();
}

# returns hash of name (from synonyms or as locus id) -> list of locusIds
sub getOrfIdsByName($@) {
    my $self = shift;
    my @names = @_;
    my $hash = &queryHashList( "SELECT DISTINCT s.name,s.locusId from Synonym s,Locus o,Position p,Scaffold sc"
			       . " WHERE sc.taxonomyId=" . $self->taxonomyId()
			       . " AND s.locusId=o.locusId and o.posId=p.posId AND p.scaffoldId=sc.scaffoldId"
			       . " AND s.name IN (" . join(",",map("'$_'",@names)) . ")" );
    my @numeric = grep m/^[0-9]+/, @names;
    if (scalar @numeric > 0) {
	map { addToHashList($hash, $_, $_) } &queryList("SELECT o.locusId FROM Locus o,Position p,Scaffold sc"
							. " WHERE sc.taxonomyId=" . $self->taxonomyId()
							. " AND o.locusId IN (" . join(",",@numeric) . ")"
							. " AND o.posId=p.posId AND p.scaffoldId=sc.scaffoldId");
    }
    return($hash);
}

sub getGenomeMap{
    return map {$_->taxonomyId(),$_} &getGenomeList();
}

sub getGenomeList{
    my %genomes = map {$_->[0],$_->[1]} @{&query( "SELECT taxonomyId,name FROM Taxonomy" )};
    my @genomes;

    foreach my $gen ( keys %genomes ){
	push @genomes, &Genome::new( taxonomyId => $gen,
				     name => $genomes{$gen} );
    }
    return @genomes;
}

return 1;

}
