autoload("loess","modreg");

### Common operations

countif <- function(x) { return(sum(ifelse(x,1,0))); }

countPerIf <- function(x,condition=NULL) {
	if(!is.null(condition)) { x <- select(x,condition); }
	 return (sum(ifelse(x,1,0))/length(x));
}

without <- function(list,columns=list$without) {
	list2 <- list;
	for (i in columns) { list2[[i]] <- NULL; }
	return(list2);
}

withoutNA <- function(x) { return(select(x,!is.na(x))); }

findIntervalAvg <- function(x,sorted) {
	index <- findInterval(x,sorted);
	return( ifelse(x < sorted[1] | x > sorted[length(sorted)-1], index,
			rank(sorted)[ifelse(index<1,1,index)] ) );
}

reverse <- function(perm) {
	r <- 1:length(perm);
	r[perm] <- 1:length(perm);
	return(r);
}

# returns a vector of matching indices or -1 in index
findMatch <- function(queries,targets) {
	o <- order(targets); # index -> sorted index
	ro <- reverse(o);
	s <- targets[o]; # sorted values
	ns <- findInterval(queries,s); # query -> sorted index
	n <- ifelse(ns<1,-1,o[ifelse(ns<1,1,ns)]); # query -> target
	return(o=o,ro=ro,ns=ns,n=n,index=ifelse(n>=1&queries==targets[ifelse(n<1,1,n)],n,-1));
}

select <- function(X,Y) { return(split(X,Y)$"TRUE"); }

RenameColumns <- function(data, newnames) {
	d2 <- list();
	for (n in names(newnames)) {
		if(is.null(data[[n]])) warning("missing name ", n);
		d2[[ newnames[[n]] ]] <- data[[n]];
	}
	return( d2 );
}

skewness <- function(x) {
	dx <- x - mean(x);
	n <- length(x);
	m2 <- mean( dx**2 );
	k2 <- m2*n/(n-1);
	m3 <- mean( dx**3 );
	k3 <- m3 * n**2/((n-1)*(n-2));
	return(k3/k2**1.5);
}

pairIndices <- function(x) {
	n <- length(x);
	return(data.frame(i1=1:(n-1),i2=2:n));
}

pairDiff <- function(x) {
	return (x[1:(length(x)-1)] - x[2:length(x)]);
}

combineColumns <- function(x) {
	n <- length(x[,1]);
	ntot <- n*length(names(x));
	out <- data.frame(x=rep(0,ntot),type=factor(rep(names(x)[1],ntot),levels=names(x)));
	for (i in 1:length(names(x))) {
		out$x[(i-1)*n+(1:n)] <- x[,i];
		out$type[(i-1)*n+(1:n)] <- levels(out$type)[i];
	}
	return(out);
}

# average information in a vector of probabilities
avginfo <- function(x) {
	return(mean(-log(ifelse(x<1e-8,1,x)) * x));
}

certainty <- function(x) {
	return(2*mean( abs(x - .5) ));
}

info <- function(x) { y <- x/sum(x); return(sum(-log(ifelse(y<1e-8,1,y))*y)); }

combineP <- function(p1, p2) {
	top <- p1/(1-p1) * p2/(1-p2);
	pm <- pmax(p1,p2);
	return(ifelse(pm >= .99999, pm, top/(1+top)));
}

logodds <- function(p) { return(log(p/(1-p))); }

logodds2p <- function(x) { return (exp(x)/(1+exp(x))); }

cut4 <- function(x,...) { return(cut(x,breaks=quantile(x),include.lowest=TRUE,...)); }
cutN <- function(x,n=8,...) { return(cut(x,breaks=unique(quantile(x,seq(0,1,1/n))),include.lowest=TRUE,...)); }

mi <- function(x,y,breaks) {
	rangeB <- range(breaks);
	# force to within range
	x1 <- pmin(pmax(x,rangeB[1]),rangeB[2]);
	y1 <- pmin(pmax(y,rangeB[1]),rangeB[2]);
	factorX <- cut(x1,breaks=c(rangeB[1]-1,breaks));
	factorY <- cut(y1,breaks=c(rangeB[1]-1,breaks));
	countX <- table(factorX);
	countY <- table(factorY);
	countXY <- table(countX,countY);
	return( info(as.vector(countX))+info(as.vector(countY))-info(as.vector(countXY)) );
}

# compute a percent >= vs. %(combined rank) plot, like for the Kolmogorov-Smirnov test
ksRanks <- function(list) {
	join <- c();
	for (name in names(list)) { join <- c(join, list[[name]]); }
	ranksJ <- floor(rank(join)); # need integers for tabulate

	out <- list();
	out$per = (1:length(join))/length(join); # percentile of the union
	index <- 0;
	for (name in names(list)) {
		ranksN <- ranksJ[index + (1:length(list[[name]]))];
		countRanks <- cumsum(tabulate(floor(ranksN),nbins=length(ranksJ)));
		out[[name]] <- countRanks/length(list[[name]]);
		index <- index + length(list[[name]]);
	}
	return(data.frame(out));
}

DratioByCutoff <- function(values, score, default=score>.5,
			minS=10, maxS=min(countif(default),countif(!default)), step=40) {
	HoverL <- countif(default)/countif(!default);
	Dall <- ks.test(select(values,!is.na(values) & default),select(values,!is.na(values) & !default))$statistic;
	sorted <- sort(score);
	Dshort <- vector();
	Dlocal <- vector();
	n <- vector();
	x <- minS;
	while (x + step < maxS && HoverL*(x+step) < maxS) {
		cutoffLow <- sorted[x];
		cutoffHigh <- sorted[length(sorted)-HoverL*x];
		cutoffL2 <- sorted[x+step];
		cutoffH2 <- sorted[length(sorted)-HoverL*(x+step)];
		Dshort[length(Dshort)+1] <- ks.test(select(values,!is.na(values) & score <= cutoffLow),
							select(values,!is.na(values) & score >= cutoffHigh))$statistic;
		Dlocal[length(Dlocal)+1] <- ks.test(select(values,!is.na(values) & score<=cutoffL2 & score>=cutoffLow),
						select(values,!is.na(values) & score<=cutoffHigh & score>=cutoffH2)
						)$statistic;
		n[length(n)+1] <- x;
		x <- x + step;
	}
	Dshort[length(Dshort)+1] <- Dall;
	Dlocal[length(Dlocal)+1] <- Dlocal[length(Dlocal)];
	n[length(n)+1] <- min(countif(default),countif(!default));
	return(data.frame(f=n/max(n),n=n,Dshort=Dshort,Dratio=Dall/Dshort,Dlocal=Dlocal));
}

# mutual information after splitting at the cutoff, versus the pooled set
# not sure this is right
miSplit <- function(x,y,cutoff) {
	xL <- sum(ifelse(x<cutoff,1,0));
	xH <- sum(ifelse(x>=cutoff,1,0));
	yL <- sum(ifelse(y<cutoff,1,0));
	yH <- sum(ifelse(y>=cutoff,1,0));
	return( info(c(xL,xH)+info(c(yL,yH)-info(c(xL+yL,xH+yH)))) );
}

# partial correlation between v1&v2, controlling for v3
pcor <- function(v1, v2, v3,spearman=FALSE) {
	if (spearman) { v1 <- rank(v1); v2 <- rank(v2); v3 <- rank(v3); }
        c12 <- cor(v1, v2)
        c23 <- cor(v2, v3)
        c13 <- cor(v1, v3)
	r <- (c12-(c13*c23))/(sqrt(1-(c13^2)) * sqrt(1-(c23^2)));
	df <- length(v1)-3;
	t <- r/sqrt((1-r*r)/df);
	# two-tailed - commented out because way too aggressive if
	# v1,v2 and v1,v3 are correlated but v2,v3 are otherwise independent...
	# the reason seems to be that the estimates of pairwise correlations
	# skew the results...
	# p <- 2*pt(abs(t),df=df,lower=FALSE);
	return(data.frame(r=r,df=df,t=t));
}

cortest2 <- function(x1, x2, y, spearman=FALSE) {
	if (spearman) { x1 <- rank(x1); x2 <- rank(x2); y <- rank(y); }
	r1 <- cor(x1,y);
	r2 <- cor(x2,y);
	z1 <- .5*(log(1+r1)-log(1-r1))*sqrt(length(x1)-3);
	z2 <- .5*(log(1+r2)-log(1-r2))*sqrt(length(x1)-3);
	z12 <- (z1-z2)/sqrt(2); # variances add - just use sqrt(2) because they are both the same
	df <- length(x1)-2;
	dx1 <- (x1-mean(x1))/sd(x1);
	dx2 <- (x2-mean(x2))/sd(x2);
	dy <- (y-mean(y))/sd(y);
	chisq1 <- sum( (dy-r1*dx1)**2 );
	chisq2 <- sum( (dy-r2*dx2)**2 );
	f <- chisq1/chisq2;
	cortestP <- cor.test(dx1-dx2,dy);
	return(data.frame(r1=r1,r2=r2,r12=cor(rank(x1),rank(x2)),z1=z1,z2=z2,z12=z12,
			p1=2*pnorm(abs(z1),lower=FALSE),
			p2=2*pnorm(abs(z2),lower=FALSE),
			p12=2*pnorm(abs(z12),lower=FALSE),
			df=df,
			f=f,pf=2*pf(ifelse(f>1,f,1/f),df1=df,df2=df,lower=FALSE),
			rpart=cortestP$estimate,ppart=cortestP$p.value));
}

# p12 and ppart both work fine if r12=0
# With high r12, however, p12 is *very* conservative, e.g. try
# ctest2 <- cortest2Test(r1=.175,r2=.175,r12=0.9,n=748,tries=1000)
cortest2Test <- function(n=100, tries=100, r1=.3, r2=.3, r12=0) {
	out <- NULL;
	for (i in 1:tries) {
		y <- rnorm(n);
		x12 <- rnorm(n);
		x1 <- r1*y + r12*x12 + sqrt(1-r1**2-r12**2)*rnorm(n);
		x2 <- r2*y + r12*x12 + sqrt(1-r2**2-r12**2)*rnorm(n);
		row <- cortest2(x1,x2,y,spearman=TRUE);
		if(is.null(out)) { out <- row; } else { out[i,] <- row; }
	}
	return(out);
}

# warning: do NOT use with binary values
spearmans <- function(data,use="pairwise.complete.obs") {
	return(cor(apply(data,2,rank),use=use));
}

corNoNa <- function(x,y, spearman=FALSE) {
	b <- !(is.na(x) | is.na(y));
	if(countif(b) < 2) { return(NA); }
	if (spearman) {
		return(cor( rank( select(x,b) ), rank( select(y,b) ) ));
	}
	#else
	return(cor( select(x,b), select(y,b) ));
}

corNoNaSelect <- function(x,y, cond, spearman=FALSE) {
	b <- cond & !is.na(x) & !is.na(y);
	if (countif(b) < 2) { return(NA); }
	if (spearman) { return( cor(rank(select(x,b)),rank(select(y,b))) ); }
	return (cor(select(x,b),select(y,b)));
}

pairCor <- function(dframe, spearman=FALSE) {
	nVar <- length(names(dframe));
	nVar2 <- nVar * (nVar-1)/2;
	nF <- as.factor(names(dframe));
	out <- data.frame(name1=rep(nF[1],nVar2), name2=rep(nF[1],nVar2), cor=rep(0,nVar2),
				i=rep(0,nVar2), j=rep(0,nVar2));
	square <- squarify(data.frame(nF=nF));
	out$i <- square$i;
	out$j <- square$j;
	out$name1 <- nF[out$i];
	out$name2 <- nF[out$j];
	for (k in 1:length(out$i)) {
		out$cor[k] <- corNoNa(dframe[,out$i[k]],dframe[,out$j[k]],spearman=spearman);
	}
	return(out);
}

squarify <- function(data) {
	nVar <- length(data[,1]);
	nVar2 <- nVar * (nVar-1)/2;
	out <- list();
	out$i = rep(0,nVar2);
	out$j = rep(0,nVar2);
	for (name in names(data)) {
		out[[paste(name,".1",sep="")]] = rep(data[1,name], nVar2);
		out[[paste(name,".2",sep="")]] = rep(data[1,name], nVar2);
	}
	out <- data.frame(out);
	index <- 1;
	for (i in 1:(nVar-1)) {
		out$i[index:(index+nVar-1-i)] = i;
		out$j[index:(index+nVar-1-i)] = (i+1):(i+1+nVar-1-i); # == (i+1):nVar
		index <- index + nVar-i;
	}
	for (name in names(data)) {
		out[,paste(name,".1",sep="")] = data[out$i,name];
		out[,paste(name,".2",sep="")] = data[out$j,name];
	}
	return(out);
}

countSimilarity <- function(square,names1,names2) {
	code <- rep(0, length(square[,1]));
	for (i in 1:length(names1)) {
		code <- ifelse(code == (i-1) & square[,names1[i]]==square[,names2[i]], code+1,code);
	}
	return(code);
}

countSimilarPairs <- function(x,names) {
	n <- length(x[,1]);
	code <- rep(0, n**2);
	for (z in 1:length(names)) {
		name <- names[z];
		for (i in 1:n) {
			index <- (1:n)+(i-1)*n;
			code[index] <- code[index] + ifelse(code[index] == (z-1) & x[i,name] == x[,name], 1,0);
		}
	}
	return(code);
}

# log(likelihood ratio) of normal deviates x and y given correlation r1 over correlation r2
# works on vectors of x,y too
logpRatio <- function(x,y,r1,r2=0) {
	return ( -.5*(x**2-y**2-2*r1*x*y)/(1-r1**2)+.5*(x**2-y**2-2*r2*x*y)/(1-r2**2)
		- .5*log((1-r1**2)/(1-r2**2)) );
}

logpRatioUnnorm <- function(x,y,r1,r2=0) {
	xNa <- select(x,!is.na(x));
	yNa <- select(y,!is.na(y));
	x1 <- (x-mean(xNa))/sd(xNa);
	y1 <- (y-mean(yNa))/sd(yNa);
	return(logpRatio(x1,y1,r1,r2));
}

### Common I/O
writeDelim <- function(table,file) {
	write.table(table,file,sep="\t",quote=FALSE,row.names=FALSE);
}


### Graphics utilities
hline <- function(y,col="grey",lty=2,lwd=1) {
	lines(c(-1e5,1e-40,1e5),c(y,y,y),col=col,lty=lty,lwd=lwd);
}

vline <- function(x,col="grey",lty=2,lwd=1) {
	lines(c(x,x,x),c(-1e5,1e-40,1e5),col=col,lty=lty,lwd=lwd);
}

eqline <- function(col="grey",lty=2,lwd=1) {
	x <- 10**(-25:25);
	lines(c(-rev(x),x),c(-rev(x),x),col=col,lty=lty,lwd=lwd);
}


defaultBreaks <- function(x,n=NULL) {
	if (is.null(n)) { n <- max(1,round(sqrt(length(x)) )); }
	return( min(x) + (0:n)*(max(x)-min(x))/n );
}

linePairs <- segments;

function(x1,y1,x2,y2,...) {
	x1 <- x1 + 0*(x1+y1+x2+y2);
	y1 <- y1 + 0*(x1+y1+x2+y2);
	x2 <- x2 + 0*(x1+y1+x2+y2);
	y2 <- y2 + 0*(x1+y1+x2+y2);
	i <- 0:(3*length(x1)-1);
	index <- 1+ifelse(i %% 3 == 0, i %/% 3, ifelse(i %% 3 == 1, length(x1)+(i%/%3), NA));
	lines(c(x1,x2)[index],c(y1,y2)[index],...);
}

sortedLines <- function(x, y, plot=FALSE, ...) {
	o <- order(x);
	if (plot) plot(x[o], y[o], type="l", ...) else lines(x[o], y[o], ...);
}

CompareDensitySelects <- function(values, list, # of criteria (1 if included in the list)
	labels=names(list),col=1:length(names(list)),lty=1:length(names(list)),
	xlim=range(values),
	plot=TRUE,
	xlab="Values",ylab="Density",xl=xlim[1],adjust=1,...) {
	dlist <- list();
	maxd <- 0;
	for (name in names(list)) {
		dlist[[name]] <- density(pmin(xlim[2],pmax(xlim[1],select(values,list[[name]]))),
						from=xlim[1],to=xlim[2],adjust=adjust);
		maxd <- max(maxd,dlist[[name]]$y); # not parallel
	}
	if (plot) {
		plot(1e-6,1e-6,col=0,xlim=xlim,ylim=c(1e-4,maxd),xlab=xlab,ylab=ylab,...);
		for (i in 1:length(names(list))) {
			name <- names(list)[i];
			lines(dlist[[name]]$x,dlist[[name]]$y,col=col[i],lty=lty[i]);
		}
		legend(xl,maxd,labels,col=col,lty=lty);
	}
	else { return(dlist); }
}

CompareDensityRatio <- function(values, truefalse, plot=TRUE, xlim=range(values),
		xlab="Values", ylab="pTrue", ylim=c(0,1), expand=0,adjust=1,...) {
	d1 <- density(pmin(xlim[2],pmax(xlim[1],select(values,truefalse))),
			from=xlim[1]-expand,to=xlim[2]+expand,adjust=adjust);
	d2 <- density(pmin(xlim[2],pmax(xlim[1],select(values,!truefalse))),
			from=xlim[1]-expand,to=xlim[2]+expand,adjust=adjust);
	if(plot) {
		plot(d1$x,d1$y/(d1$y+d2$y),type="l",xlab=xlab,ylab=ylab,xlim=xlim,ylim=ylim,...);
		hline(.5);
	} else {
		return(data.frame(x=d1$x,dT=d1$y,dF=d2$y,pTrue=d1$y/(d1$y+d2$y)));
	}
	
}

CompareDensities <- function(list,labels=names(list),xlim=range(unlist(list)),ylim=c(0,3),
		col=1:length(labels), lty=1:length(labels), lwd=rep(1,length(labels)),
		legendX=mean(xlim),legendY=ylim[2],main="",xlab="",ylab="") {
	for (i in 1:length(labels)) {
		d <- density(list[[names(list)[i]]],from=xlim[1],to=xlim[2]);
		if(i==1) {
			plot(d$x,d$y,type="l",col=col[i],lty=lty[i],lwd=lwd[i],
				main=main,xlab=xlab,ylab=ylab,xlim=xlim,ylim=ylim);
		} else {
			lines(d$x,d$y,col=col[i],lty=lty[i],lwd=lwd[i]);
		}
	}
	legend(legendX,legendY,labels,col=col,lty=lty,lwd=lwd);
}

CompareCountsList <- function(list,breaks,
				xlab="Range",ylab="Count",
				legendX=mean(breaks),legendY=-1,
				labels=names(list),
				log="",
				xlim=range(breaks),...) {
	maxlen <- 1;
	minB <- min(breaks);
	maxB <- max(breaks);
	for (name in names(list)) {
		x <- list[[name]];
		if (min(x) < minB) { cat("Warning:",name,"truncated on left\n"); }
		if (max(x) > maxB) { cat("Warning:",name,"truncated on right\n"); }
		maxlen <- max(maxlen, length(x));
	}
	if(legendY<0) { legendY <- maxlen; }
	logy <- log=="y" || log=="xy";
	plot(c(minB,maxB),c(ifelse(logy,.5,0),maxlen),pch=0,col=0,
			xlab=xlab,ylab=ylab,xlim=xlim,log=log,...);
	for (i in 1:length(names(list))) {
		name <- names(list)[i];
		x <- pmin(pmax(list[[name]],minB),maxB);
		hx <- hist(x,breaks=breaks,plot=FALSE);
		if(logy) { counts <- ifelse(hx$counts==0,.5,hx$counts); } else { counts <- hx$counts; }
		points(hx$mids,counts,pch=i,col=i);
		lines(hx$mids,counts,lty=i,col=i);
	}
	legend(legendX,legendY,labels,
		pch=1:length(names(list)),
		col=1:length(names(list)),
		lty=1:length(names(list)));
}

CompareHistogramList <- function(list,breaks,
				xlab="Range",ylab="Proportion",
				labels=names(list),
				log="", minY=ifelse(log=="y" || log=="xy",0.1, 0),
				maxlen=max(sapply(list,length)),
				ylim=c(minY/maxlen,1),
				xlim=range(breaks),
				legendX=if(log=="x" || log=="xy") exp(mean(log(xlim))) else mean(xlim),
				returnP=FALSE,plot=TRUE,warn=TRUE,
				legendY=max(ylim),cex=1,
				col=1:length(list),
				lty=1:length(list),
				pch=1:length(list),
				lwd=NULL,
				...) {
	minB <- min(breaks);
	maxB <- max(breaks);
	for (name in names(list)) {
		x <- list[[name]];
		x <- select(x,!is.na(x));
		if (warn) {
			if (min(x) < minB) { cat("Warning:",name,"truncated on left\n"); }
			if (max(x) > maxB) { cat("Warning:",name,"truncated on right\n"); }
		}
	}
	if (plot) plot(c(1/maxlen,1),c(1/maxlen,1),pch=0,col=0,lwd=1,
			xlab=xlab,ylab=ylab,log=log,xlim=xlim,ylim=ylim,...);
	outlist <- list();
	for (i in 1:length(names(list))) {
		name <- names(list)[i];
		x <- pmin(pmax(list[[name]],minB),maxB);
		x <- select(x,!is.na(x));
		hx <- hist(x,breaks=breaks,plot=FALSE);
		hxAdj <- ifelse(hx$counts==0,minY,hx$counts)/length(x);
		if (plot) {
			points(hx$mids,hxAdj,pch=pch[i],col=col[i]);
			lines(hx$mids,hxAdj,lty=lty[i],col=col[i],lwd=if(is.null(lwd)) 1 else lwd[i]);
		}
		if (returnP) {
			if (i==1) outlist[["mids"]] <- hx$mids;
			outlist[[name]] <- hxAdj;
		}
	}
	if (plot) legend(legendX,legendY,labels,cex=cex,
			pch=pch, col=col, lty=lty,
			lwd=if(is.null(lwd)) 1 else lwd);
	if(returnP) { return(data.frame(outlist)); }
}

CompareHistograms <- function(x,y,breaks,
				distr1="x",distr2="y",
				xlab="Range",ylab="Proportion",
				legendX=mean(breaks),legendY=0.9,...) {
	minB <- min(breaks);
	maxB <- max(breaks);
	if (min(x) < minB) { cat("Warning: first argument truncated on left\n"); }
	if (max(x) > maxB) { cat("Warning: first argument truncated on right\n"); }
	if (min(y) < minB) { cat("Warning: second argument truncated on left\n"); }
	if (max(y) > maxB) { cat("Warning: second argument truncated on right\n"); }
	x <- pmin(pmax(x,minB),maxB);
	y <- pmin(pmax(y,minB),maxB);
	hx <- hist(x,breaks=breaks,plot=FALSE);
	hy <- hist(y,breaks=breaks,plot=FALSE);
	plot(hx$mids,hx$counts/length(x),
		xlim=range(breaks),
		ylim=c(0,max(max(hx$counts)/length(x),max(hy$counts)/length(y))),
		xlab=xlab,ylab=ylab,...);
	lines(hx$mids,hx$counts/length(x));
	points(hy$mids,hy$counts/length(y),pch=2);
	lines(hy$mids,hy$counts/length(y),lty=2);
	legend(legendX,legendY,c(distr1,distr2),pch=c(1,2),lty=c(1,2));
}

CompareLogHistograms <- function(x,y,breaks,
				distr1="x",distr2="y",
				xlab="Range",ylab="Proportion",
				legendX=mean(breaks),legendY=0.9,...) {
	minB <- min(breaks);
	maxB <- max(breaks);
	if (min(x) < minB) { cat("Warning: first argument truncated on left\n"); }
	if (max(x) > maxB) { cat("Warning: first argument truncated on right\n"); }
	if (min(y) < minB) { cat("Warning: second argument truncated on left\n"); }
	if (max(y) > maxB) { cat("Warning: second argument truncated on right\n"); }
	x <- pmin(pmax(x,minB),maxB);
	y <- pmin(pmax(y,minB),maxB);
	hx <- hist(x,breaks=breaks,plot=FALSE);
	hy <- hist(y,breaks=breaks,plot=FALSE);
	plot(hx$mids,ifelse(hx$counts==0,.1,hx$counts)/length(x),
		xlim=range(breaks),
		ylim=c(.1/max(length(x),length(y)),1),
		xlab=xlab,ylab=ylab,log="y",...);
	lines(hx$mids,ifelse(hx$counts==0,.1,hx$counts)/length(x));
	points(hy$mids,ifelse(hy$counts==0,.1,hy$counts)/length(y),pch=2);
	lines(hy$mids,ifelse(hy$counts==0,.1,hy$counts)/length(y),lty=2);
	legend(legendX,legendY,c(distr1,distr2),pch=c(1,2),lty=c(1,2));
}

densityPlot <- function(x, y, nx=NULL, ny=NULL,
			xbreaks=defaultBreaks(x,n=nx), ybreaks=defaultBreaks(y,n=ny),
			nz=30, logZ=FALSE, byBreaks=FALSE,
			xlab="",ylab="") {
	nXY <- xtabs(~x+y, data.frame(x=findInterval(x,xbreaks), y=findInterval(y,ybreaks)));
	rectX <- row(nXY);
	rectY <- col(nXY);
	maxXY <- max(nXY);
	if (logZ) {
		intensity <- log(1+nXY)/log(1+maxXY);
		colors <- grey(1 - intensity);
	} else {
		intensity <- 1 + round(nz*nXY/maxXY);
		colors <- ifelse(nXY==0, rgb(1,1,1), grey(1 - intensity/(nz+1)));
	}
	if (byBreaks) {
		plot(range(row(nXY)),range(col(nXY)),col=0,xlab=xlab,ylab=ylab);
		rect(c(row(nXY)),c(col(nXY)),
			c(row(nXY)+1),c(col(nXY)+1),
		col=c(colors), lty=0 );
	} else {
		plot(range(x),range(y),col=0,xlab=xlab,ylab=ylab);
		rect(xbreaks[c(row(nXY))],ybreaks[c(col(nXY))],
			xbreaks[c(row(nXY)+1)],ybreaks[c(col(nXY)+1)],
		col=c(colors), lty=0 );
	}
	return(nXY);
}

### Receiver Operating Curves

# a more accurate variant
makeaoc <- function(s,d=NULL,bigger=TRUE) {
	if (is.null(d)) { d <- s$"FALSE"; s <- s$"TRUE"; }
	if (!bigger) { s <- -s; d <- -d; }
	n <- 0*s;
	for(i in 1:length(s)) { n[i] <- countPerIf(s[i]>d) + .5*countPerIf(s[i]==d); };
	return(mean(n));
}

# default is bigger is better
AOC <- function(valuesT,valuesF=NULL, truth=NULL, bigger=TRUE) {
	if (is.null(valuesF)) {
		valuesF <- select(valuesT,!truth);
		valuesT <- select(valuesT,truth);
	}
	if(!bigger) {
		valuesT <- -valuesT;
		valuesF <- -valuesF;
	}
	cutoff <- sort(unique(valuesT));
	cutoff <- c(min(valuesT,valuesF)-1, cutoff, max(valuesT,valuesF)+1);
	sens <- spec <- vector(length=length(cutoff));
	for (i in 1:length(cutoff)) {
		sens[i] <- countif(valuesT >= cutoff[i]);
		spec[i] <- countif(valuesF < cutoff[i]);
	}
	sens <- sens / length(valuesT);
	spec <- spec / length(valuesF);
	acc <- (sens+spec)/2;
	AOC <- sum( (sens[1:(length(sens)-1)]-sens[2:length(sens)]) *
			(spec[1:(length(spec)-1)]+spec[2:length(spec)])/2 );
	if(!bigger) {
		cutoff <- -cutoff;
	}
	return(sens=sens,spec=spec,acc=acc,cutoff=cutoff,AOC=AOC);
}



# assumes higher is better
cutoffs <- function(scoresTest, scoresControl) {
	sortT <- sort(scoresTest);
	sortC <- sort(scoresControl);
	cutoffs <- sortT;
	pRatios <- 0*(1:length(sortT));
	pRatios5 <- 0*(1:length(sortT));

	for(i in 1:length(sortT)) {
		pT <- sum(ifelse(sortT >= cutoffs[i],1,0));
		pC <- sum(ifelse(sortC >= cutoffs[i],1,0));
		pRatios5[i] <- (length(pT)/length(pC)) * (pC+.5)/(pT+.5);
		pRatios[i] <- (length(pT)/length(pC)) * pC/pT;
	}
	return(data.frame(cutoff=cutoffs,pRatio5=pRatios5,pRatio=pRatios));
}

# The really smart thing is probably to do some smoothing and put a cutoff on the
# local ratio of p...
chooseCutoff <- function(scoresTest,scoresControl,p=0.05) {
	cutoffX <- cutoffs(scoresTest,scoresControl);
	indexes <- split(1:length(cutoffX$cutoff),cutoffX$pRatio5 <= p)$"TRUE";
	if (length(indexes) > 0) {
		index <- indexes[1];
	} else {
		index <- length(scoresTest);
	}
	return(data.frame(cutoff=cutoffX$cutoff[index],
		sensitivity=(length(scoresTest)-index+1)/length(scoresTest),
		specificity=sum(ifelse(scoresControl < cutoffX$cutoff[index],1,0))/length(scoresControl)
			));
}

chooseNaiveCutoff <- function(scoresTest,scoresControl,p=0.01) {
	n <- round( length(scoresControl)*(1-p) ); # n = length(scoresControl) OK
	cutoffX <- (sort(scoresControl))[n];
	return(data.frame(cutoff=cutoffX,
			sensitivity=sum(ifelse(scoresTest>cutoffX,1,0))/length(scoresTest),
			specificity=n/length(scoresControl)));
}

### Operon Prediction

gncGenomeSummary <- function(gncDir,prefix,suffix,field,p=0.01) {
	pairs <- read.delim(paste(gncDir,"all.genomes",sep="/"));
	perm <- order(pairs$TaxId1);
	genomes <- unique(sort(pairs$TaxId1));
	names <- pairs$Genome1[perm[findInterval(genomes,sort(pairs$TaxId1))]];
	thresh <- vector(length=length(genomes));	
	sensitivity <- vector(length=length(genomes));	
	specificity <- vector(length=length(genomes));
	for (i in 1:length(genomes)) {
		values <- read.delim(paste(gncDir,"/",prefix,genomes[i],suffix,sep=""));
		choose <- chooseNaiveCutoff(select(values[[field]],values$Code=="Same"),
						select(values[[field]],values$Code=="Convergent"),
						p=p);
		thresh[i] <- choose$cutoff;
		sensitivity[i] <- choose$sensitivity[1];
		specificity[i] <- choose$specificity[1];
	}
	return(data.frame(genome=genomes,name=names,thresh=thresh,sens=sensitivity,spec=specificity));
}

gncGenomeAll <- function(gncDir,prefix="gnc",suffGN=".cai",suffPhylo=".phylo",p=0.05) {
	pairs <- read.delim(paste(gncDir,"all.genomes",sep="/"));
	perm <- order(pairs$TaxId1);
	genomes <- unique(sort(pairs$TaxId1));
	names <- pairs$Genome1[perm[findInterval(genomes,sort(pairs$TaxId1))]];

	threshGN <- vector(length=length(genomes));
	sensGN <- vector(length=length(genomes));
	specGN <- vector(length=length(genomes));

	threshMI <- vector(length=length(genomes));
	sensMI <- vector(length=length(genomes));
	specMI <- vector(length=length(genomes));

	#HK or housekeeping method: pmin(N1,N2)
	threshHK <- vector(length=length(genomes));
	sensHK <- vector(length=length(genomes));
	specHK <- vector(length=length(genomes));

	#CAI method (difference squared)
	threshCAId <- vector(length=length(genomes));
	sensCAId <- vector(length=length(genomes));
	specCAId <- vector(length=length(genomes));

	#CAI method (correlation like)
	threshCAI1 <- vector(length=length(genomes));
	sensCAI1 <- vector(length=length(genomes));
	specCAI1 <- vector(length=length(genomes));

	#CAI method (likelhood)
	threshCAIll <- vector(length=length(genomes));
	sensCAIll <- vector(length=length(genomes));
	specCAIll <- vector(length=length(genomes));

	CAIcor <- vector(length=length(genomes)); #Spearman of CAI, next CAI if both available

	# strand counting - needs fixed to take origins into account!
	fPlus  <- vector(length=length(genomes));
	fSame <-  vector(length=length(genomes));

	for (i in 1:length(genomes)) {
		scores <- read.delim(paste(gncDir,"/",prefix,genomes[i],suffGN,sep=""));
		phylo <- read.delim(paste(gncDir,"/",prefix,genomes[i],suffPhylo,sep=""));

		chooseGN <- chooseNaiveCutoff(select(scores$GNScore,scores$Code=="Same"),
						select(scores$GNScore,scores$Code=="Convergent"),
						p=p);
		threshGN[i] <- chooseGN$cutoff;
		sensGN[i] <- chooseGN$sensitivity;
		specGN[i] <- chooseGN$specificity;

		chooseMI <- chooseNaiveCutoff(select(phylo$MI,scores$Code=="Same"),
						select(phylo$MI,scores$Code=="Convergent"),
						p=p);
		threshMI[i] <- chooseMI$cutoff;
		sensMI[i] <- chooseMI$sensitivity;
		specMI[i] <- chooseMI$specificity;

		HK <- pmin(phylo$N1,phylo$N2);
		chooseHK <- chooseNaiveCutoff(select(HK,scores$Code=="Same"),
						select(HK,scores$Code=="Convergent"),
						p=p);
		threshHK[i] <- chooseHK$cutoff;
		sensHK[i] <- chooseHK$sensitivity;
		specHK[i] <- chooseHK$specificity;

		fPlus[i] <- sum(ifelse(scores$Strand1=="+",1,0))/length(scores$Strand1);
		fSame[i] <- sum(ifelse(scores$Code=="Same",1,0))/length(scores$Code);

		CAIreal <- select(scores$CAI,!is.na(scores$CAI)&!is.na(scores$nCAI));
		CAIrealn <- select(scores$nCAI,!is.na(scores$CAI)&!is.na(scores$nCAI));
		# use negative because higher is better
		# 2*var(x) is the neutral expectation for (x1-x2)**2
		CAId <- -ifelse(is.na(scores$CAI)|is.na(scores$nCAI),
					2*var(CAIreal)+0*1:length(scores$CAI),
					(scores$CAI-scores$nCAI)**2);
		chooseCAId <- chooseNaiveCutoff(select(CAId,scores$Code=="Same"),
						select(CAId,scores$Code=="Convergent"),
						p=p);
		threshCAId[i] <- chooseCAId$cutoff;
		sensCAId[i] <- chooseCAId$sensitivity;
		specCAId[i] <- chooseCAId$specificity;

		CAI1 <- ifelse(is.na(scores$CAI)|is.na(scores$nCAI),
				0*1:length(scores$CAI),
				(scores$CAI-mean(CAIreal))*(scores$nCAI-mean(CAIreal)/var(CAIreal)));
		chooseCAI1 <- chooseNaiveCutoff(select(CAI1,scores$Code=="Same"),
						select(CAI1,scores$Code=="Convergent"),
						p=p);
		threshCAI1[i] <- chooseCAI1$cutoff;
		sensCAI1[i] <- chooseCAI1$sensitivity;
		specCAI1[i] <- chooseCAI1$specificity;


		CAIcor[i] <- cor(CAIreal,CAIrealn);
		r <- .8; # a hack based on E. coli
		CAIll <- ifelse(is.na(scores$CAI)|is.na(scores$nCAI),
				-2*r,
				(2*(scores$CAI-mean(CAIreal))*(scores$nCAI-mean(CAIrealn))
				- r*(scores$CAI-mean(CAIreal))**2
				- r*(scores$nCAI-mean(CAIrealn))**2)
				/(sd(CAIreal)*sd(CAIrealn)) );
		chooseCAIll <- chooseNaiveCutoff(select(CAIll,scores$Code=="Same"),
						select(CAIll,scores$Code=="Convergent"),
						p=p);
		threshCAIll[i] <- chooseCAIll$cutoff;
		sensCAIll[i] <- chooseCAIll$sensitivity;
		specCAIll[i] <- chooseCAIll$specificity;
	}
	fP2M2 <- fPlus**2 + (1-fPlus)**2;
	fSameAndOperon <- (fSame - fP2M2)/(1-fP2M2);
	fOperon <- fSameAndOperon/fSame;
	return(data.frame(genome=genomes,name=names,
			threshGN=threshGN,sensGN=sensGN,specGN=specGN,
			threshMI=threshMI,sensMI=sensMI,specMI=specMI,
			threshHK=threshHK,sensHK=sensHK,specHK=specHK,
			threshCAId=threshCAId,sensCAId=sensCAId,specCAId=specCAId,
			threshCAI1=threshCAI1,sensCAI1=sensCAI1,specCAI1=specCAI1,
			threshCAIll=threshCAIll,sensCAIll=sensCAIll,specCAIll=specCAIll,
			fPlus=fPlus,fSame=fSame,fSameAndOperon=fSameAndOperon,fOperon=fOperon));
}

# Given f(p) = (a+b*p)/(c+d*p) and a Dirichlet prior on p of p**(alpha-1)*(1-p)**(beta-1),
# and data nT, nF (true/false) based on f(p), reports the
# maximum likelihood estimate of p
pseudocountLL <- function(nT,nF,a,b,c,d,alpha=2,beta=2,plot=FALSE,debug=FALSE) {
	f <- (nT+1)/(nF+nT+2);
	naivep <- (c*f-a)/(b-d*f);
	pStart <- if(naivep > .001 && naivep < .999) naivep else (a+b*.5)/(c+d*.5);
	# the function we will minimize = -log likelihood
	func <- function(q) {
		if (q < -1e-8 || q >= 1 || a+b*q < 1e-8 || c+d*q < 1e-8 || (c-a + (d-b)*q) < 1e-8) {
			# nlm is smart enough to backtrack from illegal values
			v  <- 1e10;
			attr(v,"gradient") <- 10;
			attr(v,"hessian") <- 10;
			if(debug) { cat("Illegal",q,"\n"); }
			return(v);
		}
		v <- nT*log(a+b*q) + nF*log((c-a)+(d-b)*q) - (nT+nF)*log(c+d*q) + 
			(alpha-1)*log(q) + (beta-1)*log(1-q);
		gradient <- nT*b/(a+b*q) + nF*(d-b)/(c-a+(d-b)*q) - (nT+nF)*d/(c+d*q) +
				(alpha-1)/q - (beta-1)/(1-q);
		hessian <- -nT*b**2/(a+b*q)**2 - nF*(d-b)**2/(c-a+(d-b)*q)**2 + (nT+nF)*d**2/(c+d*q)**2 -
				(alpha-1)/q**2 - (beta-1)/(1-q)**2;
		# we want to maximize; nlm minimizes
		v <- -v;
		attr(v,"gradient") <- -gradient;
		attr(v,"hessian") <- -hessian;
		return(v);
	}
	opt <- nlm(func,pStart,hessian=TRUE,check.analyticals=FALSE);
	if (plot) {
		cat("Estimate",opt$estimate,"Start",pStart,"MaxLL",-opt$minimum,
			"Gradient",-opt$gradient,"Hessian",-opt$hessian,
			"Iter",opt$iter,"Code",opt$code,"\n");
		oldpar <- par(no.readonly=TRUE);
		on.exit(par(oldpar));
		par(mfrow=c(1,2));
		p=(1:299)/300
		dll <- nT*b/(a+b*p) + nF*(d-b)/(c-a+(d-b)*p) - (nT+nF)*d/(c+d*p) +
			(alpha-1)/p - (beta-1)/(1-p);
		plot(p,dll,ylim=c(-20,20),type="b");
		vline(naivep,lty=1);
		vline(opt$estimate,lty=1,col="red");
		vline(0);
		vline(1);
		hline(0,lty=1);
		plot(p,(a+b*p)/(c+d*p),type="l",ylim=c(0,1),lwd=2);
		vline(0);vline(1);hline(0);hline(1);vline(naivep,lty=1);hline(f,lty=1);eqline();
		vline(opt$estimate,lty=1,col="red");
	} else {
		return(list(p=opt$estimate,code=opt$code,iter=opt$iter,gradient=-opt$gradient));
	}
}

OperonsData <- function(genome,name="",gncDir="operons/out",
				prefix="gnc",suffGN=".cai",suffPhylo=".phylo") {
	gcN <- read.delim(paste(gncDir,"/",prefix,genome,suffGN,sep=""));
	gcP <- read.delim(paste(gncDir,"/",prefix,genome,suffPhylo,sep=""));
	if (is.null(gcN) || is.null(gcP) || length(gcN$Code) != length(gcP$Code)
		|| is.null(gcN$Scaffold)) {
		cat("Skipping ",genome," ",name,"\n");
		return(NULL);
	}

	splits <- split(cbind(gcN,index=1:nrow(gcN)),gcN$Scaffold);
	scaffolds <- names(splits);
	starts <- termini <- list();
	allForward <- allForwardplus <- allSumBias <- rep(NA,nrow(gcN));
	for (s in names(splits)) {
		gcNS <- splits[[s]];
		ns <- 1:nrow(gcNS);
		sumBias <- cumsum(ifelse(gcNS$Strand1=="+",1,-1));
		terminus <- select(1:nrow(gcNS),sumBias==max(sumBias))[1];
		rotatedN <- if(terminus==nrow(gcNS)) ns else c((terminus+1):nrow(gcNS),1:terminus);
		rotatedS <- gcNS$Strand1[rotatedN];
		sumRotBias <- cumsum(ifelse(rotatedS=="+",1,-1));
		start <- select(rotatedN,sumRotBias==min(sumRotBias))[1];
		forwardplus <- if(terminus >= start) (ns > start & ns <= terminus) else
				(ns > start | ns <= terminus);
		forward <- (gcNS$Strand1==ifelse(forwardplus,"+","-"));

		starts[[s]] <- gcNS$index[start];
		termini[[s]] <- gcNS$index[terminus];
		allForward[gcNS$index] <- forward;
		allForwardplus[gcNS$index] <- forwardplus;
		allSumBias[gcNS$index] <- sumBias;
	}

	fPlus <- countPerIf(allForward); # a bit of a misnomer, changing from "forward" to "plus"
	fP2M2 <- fPlus**2 + (1-fPlus)**2;
	fSame <- sum(ifelse(gcN$Code=="Same",1,0))/length(gcN$Code);
	fOperonNaive <- (1 - fP2M2/fSame)/(1 - fP2M2);
	pSameLeading <- countPerIf(select(gcN$Code=="Same",allForward));
	pSameLagging <- countPerIf(select(gcN$Code=="Same",!allForward));
	fOperon <- OperonFreqStrandBias(pSameLeading,pSameLagging);

	CAIdsqr <- (gcN$CAI-gcN$nCAI)**2;
	CAIdsqr <- ifelse(is.na(CAIdsqr),median(select(CAIdsqr,!is.na(CAIdsqr))),CAIdsqr);
	CAIdxdy <- (gcN$CAI-mean(select(gcN$CAI,!is.na(gcN$CAI))))*
			(gcN$nCAI-mean(select(gcN$nCAI,!is.na(gcN$nCAI))));
	CAIdxdy <- ifelse(is.na(CAIdxdy),median(select(CAIdxdy,!is.na(CAIdxdy))),CAIdxdy);
	CAIrSame <- corNoNaSelect(gcN$CAI,gcN$nCAI, gcN$Code=="Same");
	CAIrNSame <- corNoNaSelect(gcN$CAI,gcN$nCAI, gcN$Code!="Same");

	CAIp <- ifelse(is.na(gcN$CAI),median(select(gcN$CAI,!is.na(gcN$CAI))), gcN$CAI);
	nCAIp <- ifelse(is.na(gcN$nCAI),median(select(gcN$nCAI,!is.na(gcN$nCAI))), gcN$nCAI);
	CAIrSameR <- corNoNaSelect(rank(CAIp),rank(nCAIp), gcN$Code=="Same");
	CAIrNSameR <- corNoNaSelect(rank(CAIp),rank(nCAIp), gcN$Code!="Same");
	CAIlogpratio <- logpRatioUnnorm(CAIp,nCAIp,CAIrSameR,CAIrNSameR);
	CAIdxdyR <- (rank(CAIp)-mean(rank(CAIp)))*(rank(nCAIp)-mean(rank(nCAIp)))/mean(rank(CAIp))**2;
	CAIdsqrR <- (rank(CAIp)-rank(nCAIp))**2/mean(rank(CAIp))**2;

	# report CAI bias...
	plus <- gcN$Strand1=="+";
	caiPlus <- select(gcN$CAI, !is.na(gcN$CAI) & plus);
	caiMinus <- select(gcN$CAI, !is.na(gcN$CAI) & !plus);
	caiForward <- select(gcN$CAI, !is.na(gcN$CAI) & allForward);
	caiReverse <- select(gcN$CAI, !is.na(gcN$CAI) & !allForward);
#	firstHalf <- (1:length(gcN$CAI)) < length(gcN$CAI)/2;
#	caiFirstHalf <- select(gcN$CAI, !is.na(gcN$CAI) & firstHalf);
#	caiSecondHalf <- select(gcN$CAI, !is.na(gcN$CAI) & !firstHalf);
#	caiStartHalf <- select(gcN$CAI, !is.na(gcN$CAI) & forwardplus);
#	caiTerminusHalf <- select(gcN$CAI, !is.na(gcN$CAI) & !forwardplus);
#	# also compute forward based on terminus at 0, just in case
#	caiForward0 <- select(gcN$CAI, !is.na(gcN$CAI) & ((firstHalf & plus) | (!firstHalf & !plus)));
#	caiReverse0 <- select(gcN$CAI, !is.na(gcN$CAI) & !((firstHalf & plus) | (!firstHalf & !plus)));

	return(list(genome=genome,name=name,
		gc=cbind(gcN,MI=gcP$MI, HK=pmin(gcP$N1,gcP$N2),
			CAIdsqr=CAIdsqr,CAIdxdy=CAIdxdy,
			CAIlogpratio=CAIlogpratio,CAIdxdyR=CAIdxdyR,CAIdsqrR=CAIdsqrR,
			COGsim=COGSimilarity(gcN$COGfun,gcN$nCOGfun),
			N1=gcP$N1, N2=gcP$N2,
			forward=allForward,forwardplus=allForwardplus,sumBias=allSumBias),
		fPlus=fPlus,fP2M2=fP2M2,fSame=fSame,fOperon=fOperon,fOperonNaive=fOperonNaive,
		scaffolds=scaffolds,starts=starts,termini=termini,
		pCAIPlusMinus = wilcox.test(caiPlus,caiMinus)$p.value,
		pCAIForwardReverse = wilcox.test(caiForward,caiReverse)$p.value,
#		pCAIFirstSecondHalves = wilcox.test(caiFirstHalf,caiSecondHalf)$p.value,
#		pCAIStartTerminusHalves = wilcox.test(caiStartHalf,caiTerminusHalf)$p.value,
#		pCAIForwardReverse0 = wilcox.test(caiForward0,caiReverse0)$p.value,
		CAIrSame=CAIrSame, CAIrNSame=CAIrNSame,
		CAIrSameR=CAIrSameR, CAIrNSameR=CAIrNSameR ))
}

BiasPlot <- function(d,plotDetails=TRUE) {
	plot(1:length(d$gc$Strand1),d$gc$sumBias/length(d$gc$Strand1),type="l",
		xlab=paste("Gene # in",d$genome),ylab="Cumulative Strand Bias",main=d$name);
	for (scaffold in d$scaffolds) {
		vline(d$starts[[scaffold]],col="green",lty=1);
		vline(d$termini[[scaffold]],col="red",lty=1);
	}
	for (scaffold in d$scaffolds) {
		vline(min(select(1:nrow(d$gc),d$gc$Scaffold==scaffold)),lwd=2);
		vline(max(select(1:nrow(d$gc),d$gc$Scaffold==scaffold)),lwd=2);
	}
	hline(0);
	if (plotDetails) {
		text(1,max(d$gc$sumBias/length(d$gc$Strand1)),
			paste("fPlus = ",format(d$fPlus,digits=3),"\n",
				"fSame = ",format(d$fSame,digits=3),"\n",
				"fSame no operons = ",format(d$fP2M2,digits=3),"\n",
				"fOperon = ",format(d$fOperon,digits=3),sep=""),
			adj=c(0,1.5));
	} else {
		text(1,max(d$gc$sumBias/length(d$gc$Strand1)),format(d$fPlus,digits=3),adj=c(0,0));
	}
}

OperonsDescribe <- function(genome,name="",log="",plot="all",sep=1,...) {
	d <- OperonsData(genome,...);
	if(is.null(d)) return();
	if (plot=="all") {
		old.par <- par(no.readonly=TRUE);
		on.exit(par(old.par));
		par(mfrow=c(3,3));
	}
	if (plot=="sep") {
		old.par <- par(no.readonly=TRUE);
		on.exit(par(old.par));
		par(mfrow=c(3,1));
	}
	if (plot=="all"||plot=="GNScore")
	  CompareHistogramList(split(d$gc$GNScore,d$gc$Code),breaks=c(0:15),xlab="log(1+GNScore)",log=log);
	if (plot=="all"||plot=="MI")
	  CompareHistogramList(split(d$gc$MI,d$gc$Code),breaks=c(-3:6)/20,xlab="MI",log=log);
	if (plot=="all"||plot=="HK")
	  CompareHistogramList(split(d$gc$HK,d$gc$Code),breaks=(0:20)*2,xlab="HK",log=log);
	if (plot=="all"||plot=="CAIdxdy")
	  CompareHistogramList(split(d$gc$CAIdxdy,d$gc$Code),breaks=(-2:5)/100,xlab="CAIdxdy",log=log);
	if (plot=="all"||plot=="CAIdsqr")
	  CompareHistogramList(split(d$gc$CAIdsqr,d$gc$Code),breaks=(0:15)/150,xlab="CAIdsqr",log=log);
	if (plot=="all"||plot=="bias") BiasPlot(d);
	gcSame <- select(d$gc,d$gc$Code=="Same");
	if (plot=="all"||plot=="sep") {
	  CompareHistogramList(split(gcSame$Sep,gcSame$GNScore>0),
		breaks=if(sep==1) (-10:20)*10 else c(-100,-50,-20,-10,0,10,20,30,40,50,75,100,150,200),
		xlab="Same Separation if has GNScore vs. not",log=log);
	  gcSameSorted <- gcSame[order(gcSame$Sep),];
	  plot(1:length(gcSameSorted$Sep),cumsum(ifelse(gcSameSorted$GNScore>0,1,-1)),type="l");
	  plot(1:length(gcSameSorted$Sep),pmin(200,pmax(-100,gcSameSorted$Sep)),type="l");
	  hline(-10); hline(0); hline(10); hline(30); hline(50);
	}
}

# returns single character codes if identical, empty strings if missing or unknown categories,
# or "0" if different
# Assumes factors as input
COGSimilarity <- function(x1,x2) {
	out <- rep("",length(x1));
	for (i in 1:length(x1)) {
		v1 <- strsplit(levels(x1)[x1[i]],"")[[1]];
		v2 <- strsplit(levels(x2)[x2[i]],"")[[1]];
		if (length(v1)==1 && length(v2)==1 && v1[1] != "-" && v2[1] != "-"
			&& v1[1] != "R" && v2[1] != "R"
			&& v1[1] != "S" && v2[1] != "S") {
			out[i] <-  if (v1[1] == v2[1]) v1[1] else "0";
		}
	}
	return(out);
}

OperonsPredict <- function(genome,name,plot="all", plotDetails=TRUE,
				control="both", # or just "Convergent" or just "Divergent"
				vars=c("GNScore","GNMinus","GNWithin","GNAll","HK","MI",
					"CAIdxdyR","CAIdsqrR"),
				postCAIdxdy=TRUE,
				useCOGsim=TRUE,
				usefOperonNaive=FALSE,
				fControlLikeOp=.1,
				data=NULL,
				formula= same ~ bbfGNScore+bbfGNWithin+bbfGNMinus+cfCOG,
				anova=TRUE,
				smooth=TRUE,
				debug=FALSE,...) {
	if (plot == "all") {
		old.par <- par(no.readonly=TRUE);
		on.exit(par(old.par));
		par(mfrow=c(3,2));
	}
	if (is.null(data)) {
		data <- OperonsData(genome,name=name,...);
	}
	if(is.null(data)) { return(NULL); }
	fOperon <- if (usefOperonNaive) data$fOperonNaive else data$fOperon;

	if (plot == "bias" || plot == "all") {
		BiasPlot(data,plotDetails=plotDetails);
	}
	if (control == "both") {
		gcSC <- data$gc;
	} else {
		gcSC <- select(data$gc,data$gc$Code==control|data$gc$Code=="Same");
	}
	gcSC <- cbind(gcSC, same=(gcSC$Code=="Same"));
	for(v in vars) {
		if(debug) { cat("BBF for",v,"\n"); }
		# Use span=1 so it doesn't blow up on the huge GNScore=0 group
		# Makes sense anyway as we expect roughly linear fits...
		bbf <- BinnedBinaryFit(gcSC[[v]], gcSC$same,smooth=smooth,plot=debug,xlab=v,span=1);
		if(debug) { cat(v,range(bbf$p),"\n"); }
		gcSC[[paste("bbf",v,sep="")]] <- log(bbf$p/(1-bbf$p));
	}

	if (useCOGsim) {
		map <- rep("Y", length(levels(gcSC$COGsim)));
		map[select(1:length(levels(gcSC$COGsim)), levels(gcSC$COGsim)=="")] <- "U";
		map[select(1:length(levels(gcSC$COGsim)), levels(gcSC$COGsim)=="0")] <- "N";
		gcSC$COGsimClass <- as.factor(map[codes(gcSC$COGsim)]);
		pCOG <- ClassFit(gcSC$COGsimClass, gcSC$same)$p[codes(gcSC$COGsimClass)];
		gcSC$cfCOG <- log(pCOG/(1-pCOG));
	}

	bbfvars <- sapply(vars, function(x) { return(paste("bbf",x,sep="")); } );
	if (useCOGsim) { bbfvars <- c(bbfvars,"cfCOG"); }

	gcm <- glm(formula,family=binomial(logit),data=gcSC);
	gcSC$pSame <- fitted(gcm);
	pCoverS <- countif(!gcSC$same)/countif(gcSC$same);

	# And now get significance values by comparing chi-squares
	sig <- NULL; # reports all the significance tests, anova, regression coefficients
	if (!is.null(bbfvars) && anova) {
		gm1 <- NULL;
		for (v in bbfvars) {
			gm <- glm(eval(parse(text=paste( "same ~ ", v, sep=""))),
					family=binomial(logit),data=gcSC);
			sig[[paste("p",v,sep="")]] <- pchisq(gm$null.deviance-gm$deviance,
								df=1, lower=FALSE);
			if(v == bbfvars[1]) { gm1 <- gm; }
		}
		if (length(bbfvars) > 1) {
			for (v in bbfvars[2:length(bbfvars)]) {
				gm <- glm(eval(parse(text=paste( "same ~ ", bbfvars[1], " + ", v,sep=""))),
						family=binomial(logit),data=gcSC);
				sig[[paste("p", bbfvars[1],v,sep="")]] <- pchisq(gm1$deviance-gm$deviance,
										df=1, lower=FALSE);
			}
		}
	}

	# And also do anova on the model (in the specified order)
	anovaOut <- if(anova) anova(gcm,test="Chisq") else NULL;
	for (v in bbfvars) {
		tmp <- as.list(gcm$coefficients)[[v]];
		if (!is.null(tmp)) {
			if(anova) sig[[paste("anova",v,sep="")]] <- anovaOut[v,"P(>|Chi|)"];
			sig[[paste("coeff",v,sep="")]] <- tmp[1];
			sig[[paste("coeffScaled",v,sep="")]] <- tmp[1]*sd(gcSC[,v]);
		}
	}
	sig[["coeffIntercept"]] <- gcm$coefficients[1];

	# Split the same pairs into high and low pSame
	# Because pSame is highly skewed, and has a lot of values around a median,
	# use quantiles to get our cutoff for doing bbfDistance
	gcSame <- select(gcSC,gcSC$Code=="Same");
	threshPSame <- quantile(gcSame$pSame,fOperon,name=FALSE);

	if (plot=="hist" || plot == "all") {
		values <- select(gcSC$pSame,gcSC$Code=="Same");
		hist(values, xlab=paste("pSame for",genome),main=genome,breaks=20);
		vline(1/(1+pCoverS),col="red");
		vline(threshPSame,col="green");
	}

	# % of Same and Convergent pairs with pSame above the threshold
	# note we use >threshPSame -- this is important if (as is often the case) there
	# is a huge clump of values in the middle with GNScore=0, GNWithin=0 -- this ensures that we
	# count those as non-operon pairs
	fHighS <- countPerIf(gcSame$pSame > threshPSame);
	fHighSC <- countPerIf(gcSC$pSame > threshPSame);
	fHighC <- countif(gcSC$pSame > threshPSame & !gcSC$same)/countif(!gcSC$same);

	# Now estimate the %operon pairs in the high and low sets, assuming the Convergent distribution
	# matches the operon distribution.
	# H means High, O means operon, g means given:
	# thresholds .95 and .1 are arbitrary - could use Dirichlet priors but with what alpha/beta??
	# also note that we use fHighS not fHighSC because only same pairs are being considered
	# now for same pairs p(!O|H) = p(H|!O)*p(!O)/p(H) = p(H|C)*p(!O)/p(H), so
	fOgH = 1 - (fHighC*(1-fOperon)/fHighS);
	fOgHUse <- pmin(fOgH,.95);
	# similarly, for same pairs p(O) = p(O|H)*p(H) + p(O|L)*p(L), so
	fOgL = (fOperon - fOgH*fHighS)/(1-fHighS);
	fOgLUse <- pmax(fOgL,.1);

	# for a given separation D, we need to get observed %high pairs fHigh = a+b*p/(c+d*p),
	# where p = P(D|O)/(P(D|O)+P(D|!O)):
	# fHigh = P(H|D) = P(H|D)/(P(H|D)+P(L|D)) = P(D|H)*P(H)/(top + same for L)
	#       = { P(D|O)*P(O|H)*P(H) + P(D|!O)*P(!O|H)*P(H) }/(top + same for L)
	#       = { p*P(O|H)*P(H) + (1-p)*P(!O|H)*P(H) }/(top + same for L)
	#       = { P(!O|H)*P(H) + p*(P(O|H)-P(!O|H))*P(H) }/(top + same for L)
	# and P(H) = fHighS since only same pairs are relevant here...
	a <- (1-fOgHUse)*fHighS;
	b <- (fOgHUse-(1-fOgHUse))*fHighS;
	c <- a + (1-fOgLUse)*(1-fHighS);
	d <- b + (fOgLUse-(1-fOgLUse))*(1-fHighS);

	# Compute pSameD, based on interpolating P(D|O)/(P(D|O)+P(D|!O)) for each group
	# Each group is based on overlapping bins (given by gSize, gSep) determined in rank-space,
	# and the p for that group is determined via Dirichlet priors and a,b,c,d
	# If our samples were large, uncontaminated, and of equal size (fHighS=.5), then p(D) = nH/(nH+nL)

	bbfD <- BinnedBinaryFit(gcSame$Sep,gcSame$pSame>threshPSame,a=a,b=b,c=c,d=d,debug=debug,
				plot=(plot=="sep"||plot=="all"),xlab="Gene Separation",log="y",main=name,
				smooth=smooth);
	if(countif(is.na(bbfD$p)) > 0 || length(bbfD$p) != length(gcSame$Sep)) {
		cat("BinnedBinaryFit failed");
		return(NULL);
	}
	pSameD <- bbfD$p;

	# And now make predictions:
	# P(O|D,X) = P(O^D^X)/{P(O^D^X)+P(!O^D^X)} = P(D|O)*P(X|O)*P(O)/(top + same for !O)
	#          = P(D|O)/P(D|!O) * P(X|O)/P(X|!O) * P(O) / (top + P(!O))
	#   where P(D|O)/P(D|!O) = pD/(1-pD) from the above scheme, and we solve
	#   P(X|S) = P(X|O)*P(O) + P(X|!O)*P(!O), plus P(X|!O)=P(X|C), to get
	#   P(X|O)/P(X|!O) = { P(X|S) - P(X|!O)*P(!O) }/{ P(O) * P(X|!O) }
	#                  = { P(X|S)/P(X|C) - P(!O) }/P(O) = { P(C)/P(S) * pSame/(1-pSame) - P(!O) }/P(O)
	# so altogether we have
	# P(O|D,X) = pD/(1-pD) * ( P(C)/P(S) * pSame/(1-pSame) - P(!O) ) / (top + P(!O))

	pLogisticRatio <- pCoverS*gcSame$pSame/(1-gcSame$pSame);
	pOpLogistic <- pmax(0.01, (pLogisticRatio - (1-fOperon))); # arbitrary floor to keep it from going -
	pOpLogistic <- pOpLogistic/(pOpLogistic + 1 - fOperon);

	# a variant where we just scale the odds ratio from the regression
	pOpLogisticM <- pLogisticRatio * fOperon/(1-fOperon);
	pOpLogisticM <- pOpLogisticM/(1+pOpLogisticM);

	# a variant where we assume some of the control pairs do look like operon pairs
	# will also reduce the subtraction problem
	# specifically, assume
	# P(X|!O) = (1-f)*P(X|C) + f*P(X|O)
	# P(X|S) = P(O)*P(X|O) + P(!O)*((1-f)*P(X|C) + f*P(X|O))
	# P(X|O) = { P(X|S) - P(!O)*(1-f)*P(X|C) } / { P(O) + f*P(!O) }
	# P(X|C) = { P(X|!O) - f*P(X|O) }/(1-f)
	# P(O|X)/P(!O|X) = P(X|O)/P(X|!O) = P(O)/P(!O) * { P(X|S) - P(!O)*(1-f)*P(X|C) }/ { P(O) + f*P(!O) }
	#					/ { (1-f)*P(X|C) + f*P(X|O) }
	# = P(O)/P(!O) * { P(X|S) - P(!O)*(1-f)*P(X|C) }/ { P(O) + f*P(!O) } / { (1-f)*P(X|C) + f*top }
	#	set indicator = P(X|S)/P(X|C) - P(!O) * (1-f)
	# = P(O)/P(!O) * indicator / ( 1-f + f*indicator )
	indicator <- pmax(.02, pLogisticRatio - (1-fOperon)*(1-fControlLikeOp));
	pllTop <- indicator / (1-fControlLikeOp + fControlLikeOp*indicator);
	pOpLogisticLike <- pllTop/(1+pllTop);

	pOpDistance <- combineP(pSameD, fOperon);
	pOp1 <- combineP(pOpLogistic,pSameD);

	# bOp1 gives the "right" #same-operon pairs. Unclear if this would be better
	# than pOp1 >= .5 as a cutoff or not
	threshOp1 <- sort(pOp1)[round((1-fOperon)*length(pOp1))];
			
	if (plot=="histOp1"||plot=="all") {
		hist(pOp1,main=name,breaks=c(0:20)/20);
		vline(0); vline(.5); vline(1); vline(threshOp1,lty=1);
	}


	# And, finally, add in CAI via binned binary fit
	# Get P(C|O)/{P(C|O)+P(C|!O)} by assuming observed set is a good sample
	# (pretend there is no contamination -- too hard to say how much there is)
	if (postCAIdxdy) {
		# want pCAI to be .5 not fOperon if no evidence
		if(debug) { cat("bbfCAI\n"); }
		bbfCAI <- BinnedBinaryFit(gcSame$CAIdxdyR-gcSame$CAIdsqrR, pOp1>threshOp1,
						fT=fOperon, smooth=smooth);
		if(debug) { cat("bbfCAI",range(bbfCAI$p),"\n"); }
	} else {
		bbfCAI <- list(p=.5);
	}
	pOp <- combineP(pOp1, bbfCAI$p);
	threshOp <- sort(pOp)[round((1-fOperon)*length(pOp))];

	if (plot=="histOp"||plot=="all") {
		hist(pOp,main=name,breaks=c(0:20)/20);
		vline(0); vline(.5); vline(1); vline(threshOp,lty=1);
	}

	preds <- cbind(gcSame,groupIndexD=bbfD$index,
			pSameD=pSameD,
			pOpLogistic=pOpLogistic,pOpDistance=pOpDistance,pOp1=pOp1, bOp1=(pOp1 > 0.5),
			pCAI=bbfCAI$p, pOp=pOp, bOp=(pOp > 0.5),
			pOp2=combineP(combineP(pOpLogisticLike,pSameD),bbfCAI$p),
			pSame=gcSame$pSame,pOpLogisticLike=pOpLogisticLike, pOpLogisticM=pOpLogisticM );

	# more data to record
	ss <- spearmans(preds[,c("pOp","pOpLogistic","pOpDistance","Sep","GNScore","GNWithin","pCAI",
					"CAIdxdyR","CAIdsqrR")]);
	ss2 <- c(ss);
	names(ss2) <- c(outer(row.names(ss),row.names(ss),paste,sep="VS"));
	quantilesSepHighOpLogistic <- quantile(select(preds$Sep,preds$pOpLogistic>.9),(1:3)/4,names=FALSE);
	ksSepLeading <- ks.test(select(preds$Sep,preds$forward),select(preds$Sep,!preds$forward));
	ksPSameLeading <- ks.test(select(preds$pSame,preds$forward),select(preds$pSame,!preds$forward));
	ksGNScoreLeading <- ks.test(select(preds$GNScore,preds$forward),select(preds$GNScore,!preds$forward));
	return(c(data, sig,
		as.list(ss2),
		list(control=control, pCoverS=pCoverS, rLogistic=cor(gcSC$pSame, ifelse(gcSC$Code=="Same",1,0)),
		gcSC=gcSC,
		fSameGN=countPerIf(select(gcSC$GNScore>0, gcSC$same)),
		fControlGN=countPerIf(select(gcSC$GNScore>0, !gcSC$same)),
		fLeadingSameGN=countPerIf(select(preds$GNScore>0, preds$forward)),
		fLaggingSameGN=countPerIf(select(preds$GNScore>0, !preds$forward)),
		fHighS=fHighS, fHighC=fHighC, fHighSC=fHighSC, fOgH=fOgH, fOgL=fOgL, a=a, b=b, c=c, d=d,
		nSame=countif(data$gc$Code=="Same"),
		nConv=countif(data$gc$Code=="Convergent"),
		nDiv=countif(data$gc$Code=="Divergent"),
		nAllLeading = countif(data$gc$forward),
		nAllLagging = countif(!data$gc$forward),
		fOpCandidateLeading = countPerIf(preds$forward),
		ksLeadingSepD = ksSepLeading$statistic,
		ksLeadingSepP = ksSepLeading$p.value,
		medianLeadingSepD = median(select(preds$Sep,preds$forward))-median(select(preds$Sep,!preds$forward)),
		medianLeadingSep = median(select(preds$Sep,preds$forward)),
		medianLaggingSep = median(select(preds$Sep,!preds$forward)),
		ksLeadingPSameD = ksPSameLeading$statistic,
		ksLeadingPSameP = ksPSameLeading$p.value,
		medianLeadingPSameD = median(select(preds$pSame,preds$forward))-median(select(preds$pSame,!preds$forward)),
		ksLeadingGNScoreD = ksGNScoreLeading$statistic,
		ksLeadingGNScoreP = ksGNScoreLeading$p.value,
		meanLeadingGNScoreD = mean(select(preds$GNScore,preds$forward))-mean(select(preds$GNScore,!preds$forward)),
		fPredOp1=countPerIf(pOp1>=.5),
		threshPSame=threshPSame,
		threshOp1=threshOp1,
		fPredOp=countPerIf(pOp>=.5),
		threshOp=threshOp,
		bbfD=bbfD,
		bbfCAI=bbfCAI,
		preds=preds,
		gcm=gcm,
		gSize=bbfD$gSize, gSep=bbfD$gSep,
		cerLogistic=certainty(pOpLogistic),
		cerDistance=certainty(pOpDistance),
		cerLD=certainty(pOp1),
		cerLDC=certainty(pOp),
		formula=paste(deparse(formula)),
		postCAIdxdy=postCAIdxdy, useCOGsim=useCOGsim, usefOperonNaive=usefOperonNaive, smooth=smooth,
		vars=vars,
		fControlLikeOp=fControlLikeOp,

		fSepMinus3=countPerIf(preds$Sep==-3),
		fSep0=countPerIf(preds$Sep==0),

		pOpLogisticSepMinus3=mean(select(preds$pOpLogistic,preds$Sep==-3)),
		pOpLogisticSep0=mean(select(preds$pOpLogistic,preds$Sep==0)),
		fOpLogisticSepCanonical=countPerIf(select(preds$pOpLogistic>.5,preds$Sep==0|preds$Sep==-3)),

		pOpSepMinus3=mean(select(preds$pOp,preds$Sep==-3)),
		pOpSep0=mean(select(preds$pOp,preds$Sep==0)),
		fOpSepCanonical=countPerIf(select(preds$pOp>.5,preds$Sep==0|preds$Sep==-3)),

		fLeadingCanonical=countPerIf(select(preds$Sep==0|preds$Sep==-3,preds$forward)),
		fLaggingCanonical=countPerIf(select(preds$Sep==0|preds$Sep==-3,!preds$forward)),

		# these aren't actually pOpDistance because they aren't corrected for fOperon
		# also note lack of smoothing...
		pOpDistanceSepMinus10=BBFPredict(-10,bbfD)$p,
		pOpDistanceSepMinus3=BBFPredict(-3,bbfD)$p,
		pOpDistanceSep0=BBFPredict(0,bbfD)$p,
		pOpDistanceSep25=BBFPredict(25,bbfD)$p,
		pOpDistanceSep50=BBFPredict(50,bbfD)$p,
		pOpDistanceSep100=BBFPredict(100,bbfD)$p,
		pOpDistanceSep150=BBFPredict(150,bbfD)$p,
		pOpDistanceSep200=BBFPredict(200,bbfD)$p,
		pOpDistanceSep300=BBFPredict(300,bbfD)$p,

		# Moreno-Hagelsieb and Collado-Vides propose that -20 to 30 is always roughly operon pairs!
		fMoreno=countPerIf(preds$Sep >= -20 & preds$Sep <= 30),
		fMorenoGN=countPerIf(select(preds$Sep >= -20 & preds$Sep <= 30,preds$GNScore>0)),
		fMorenoGN0=countPerIf(select(preds$Sep >= -20 & preds$Sep <= 30,preds$GNScore==0)),

		quartile1SepHighOpLogistic = quantilesSepHighOpLogistic[1],
		medianSepHighOpLogistic = quantilesSepHighOpLogistic[2],
		quartile3SepHighOpLogistic = quantilesSepHighOpLogistic[3],
		scaffoldsS=paste(data$scaffolds,collapse=","),
		startsS=paste(data$starts,collapse=","),
		terminiS=paste(data$termini,collapse=","),
		without=c("without","gc","gcSC","preds","gcm","bbfD","bbfCAI","vars",
				"scaffolds","starts","termini")
	)));

	# The old way with density ratios:
	#	dRatio <- CompareDensityRatio(gcSame$Sep,(gcSame$pSame>=1/(1+pCoverS)),
	#					xlim=c(-100,200),expand=50,plot=FALSE);
	#	pRatioX <- ( pCoverS*gcSame$pSame/(1-gcSame$pSame) - (1-fOperon) )/fOperon;
	#	index <- pmin(length(dRatio$x),pmax(1,findInterval(gcSame$Sep,dRatio$x)));
	#	top <- pmax(0,pRatioX*(dRatio$dT[index]/dRatio$dF[index])*fOperon);
	#	pOp <- ifelse(gcSame$pSame > .99999, 1, top/(top+1-fOperon));
	#	pSameHigh <- sum(ifelse(gcSame$pSame>=1/(1+pCoverS),1,0))/length(gcSame$pSame);
	#	pOgivenLow <- (fOperon - pSameHigh)/(1-pSameHigh);
	#	dFAdj <- dRatio$dF-dRatio$dT*pOgivenLow;
	#	dFAdj <- dFAdj*sum(dRatio$dT)/sum(dFAdj); # normalize
	#	dRatioAdj <- dRatio$dT/dFAdj;
	#	topAdj <- pmax(0,pRatioX*dRatioAdj[index]*fOperon);
	#	pOpAdj <- ifelse(gcSame$pSame > .99999, 1, topAdj/(topAdj+1-fOperon));
}

BinnedBinaryFit2 <- function(xT, xF, gSize=NULL, gSep=NULL, xout=c(xT,xF),
				fT=length(xT)/(length(xT)+length(xF)), # to compute abcd
				pActual=.5,
				a=NULL,b=NULL,c=NULL,d=NULL,
				smooth=TRUE,span=NULL,
				...) {
	x <- c(xT,xF);
	y <- ( 1:length(x) <= length(xT) );
	if (!is.null(a)) { fT <- NULL; } # don't override abcd if passed in
	return(BinnedBinaryFit(x,y, gSize=gSize, gSep=gSep, xout=xout,
				fT=fT, pActual=pActual, a=a, b=b, c=c, d=d, smooth=smooth, span=NULL, ...));
}

# x is numeric, y is binary, and a,b,c,d are for pseudocountLL
# returns a groups object with n(Observed F),n(Observed T), and p(Actual|x) for the group,
# with group bounds (min/med/max).
# Also, for each xout, returns  p=p(Actual) and index=which group
#
# More strictly, speaking, expects input x taken by sampling values with y=0 or 1.
# The sampling may be biased, taken according to p(y=T) = fT, as opposed to the actual
# distribution p(y=T) = pActual, and there may be noise in the category assignments,
# leading to the general form P(T observed|x) = a+b*p(T actual|x)/(c + d*p(T actual|x))
#
# Within each bin of x (60-200 overlapping bins, unless you override gSize or gSep),
# it observes nTobserved = n(X ^ y=T) and nFobserved = n(X ^ y=F), and uses pseudocountLL to estimate
# p(Actual|X) from nTobserved and nFobserved
#
# Then it interpolates between bins to get compute p(Actual|xout) for each xout
# Optionally, it uses loess (on rank(x) vs. odds ratios) to smooth this p-value
#
BinnedBinaryFit <- function(x,y,gSize=NULL,gSep=NULL,xout=x,fT=NULL,pActual=.5,a=0,b=1,c=1,d=0,
			alpha=2,beta=2,debug=FALSE,
			plot=FALSE,log="",ylim=NULL,smooth=TRUE,span=NULL,...) {
	if (!is.null(fT)) {
		# notice this is a no-op if fT==pActual
		ratio <- (pActual/(1-pActual)) * (1-fT)/fT;
		a = 0; b <- 1; c <- ratio; d <- 1-c;
	}
	if (is.null(gSize)) {
		if (length(x) < 200) { cat("Warning: BinnedBinaryFit on only",length(x),"items\n"); }
		gSize <- 2 * (length(x) %/% 40);
		gSize <- pmin(200,pmax(100,gSize));
	}
	if (is.null(gSep)) gSep <- gSize %/% 2;

	orderX <- order(x);
	xSorted <- x[orderX];
	nG <- ((length(x)-gSize) %/% gSep);
	gBeg <- 1 + gSep*(0:(nG-1));
	gEnd <- gBeg + gSize-1;
	gEnd[length(gEnd)] <- length(x); # the last group includes up to an extra gSep
	gMin <- xSorted[gBeg];
	gMax <- xSorted[gEnd];
	nT <- nF <- p <- gMed <- p <- 0*(1:length(gBeg)); 
	for (i in 1:length(p)) {
		# use all values between min,max in case of a large group with the same sep
		gx <- select(x,x >= gMin[i] & x <= gMax[i]);
		gy <- select(y,x >= gMin[i] & x <= gMax[i]);
		gMed[i] <- median(gx);
		nT[i] <- countif(gy);
		nF[i] <- length(gy) - nT[i];
		pLL <- pseudocountLL(nT[i],nF[i],a,b,c,d,alpha=alpha,beta=beta);
		if (pLL$code > 2) {
			cat("Warning: pseudocount failed","nT",nT[i],"nF",nF[i],
				"a",a,"b",b,"c",c,"d",d,"code",pLL$code,"p",pLL$p,"\n");
		}
		p[i] <- pLL$p;
		if(debug) {
			cat("group",i,"min",gMin[i],"max",gMax[i],
				"n",length(gx),"nT",nT[i],"nF",nF[i],"p",p[i],"\n");
		}
	}

	if (plot) {
		if (is.null(ylim)) { ylim <- (if(log=="y"||log=="xy") c(.001+min(p),0.5+max(c(nT,nF)))
					else c(0,1)); }
		plot(gMed,p,log=log,type="l", ylim=ylim, ...);
		for(i in 1:length(gMin)) {
			lines(c(gMin[i], gMax[i]),c(p[i],p[i]),lty=1,col="grey");
		}
		if (log=="y"||log=="xy") {
			points(gMed,.5+nT,type="l",col=3);
			points(gMed,.5+nF,type="l",col=2);
		}
	}

	bbf <- list(xSorted=xSorted,groups=data.frame(min=gMin,med=gMed,max=gMax,p=p,nT=nT,nF=nF),
			gSize=gSize,gSep=gSep,a=a,b=b,c=c,d=d,smooth=smooth,span=span);
	if (smooth && length(unique(x)) < 2) smooth = FALSE;
	if (smooth) {
		out <- BBFPredict(x, bbf); # unsmoothed b/c loess isn't set
		lo <- logodds(out$p);
		bbf$loess <- loess(lo~rank(x),span=if(is.null(span)) .5 else span);
		out <- BBFPredict(xout,bbf);
		bbf$index <- out$index;
		bbf$p <- out$p;
		bbf$pRough <- out$pRough;
		if (plot) { lines(sort(xout), out$p[order(xout)], col=3); }
	} else {
		out <- BBFPredict(xout,bbf);
		bbf$index <- out$index;
		bbf$p <- out$p;
	}
	return(bbf);
}

# interpolate because it should help in fitting separations for small genomes
# (e.g. Buchnera), and makes little difference elsewhere
# rank(median for group i) = 1 + gSep*(i-1) + gSize/2, so
# i = 1 + (rank - 1 - gSize/2)/gSep
BBFPredict <- function(xout, bbf) {
	indexScaled <- pmax(1,pmin(length(bbf$xSorted),findIntervalAvg(xout,bbf$xSorted)));
	index <- 1 + ( indexScaled - 1 - bbf$gSize/2 )/bbf$gSep;
	indexMM <- pmin(length(bbf$groups$p),pmax(1,index));
	p <- approx(1:length(bbf$groups$p),bbf$groups$p, xout=indexMM)$y;
	if(!is.null(bbf$loess)) {
		pRough <- p;
		limits <- range(bbf$loess$x);
		indexScaled <- pmin(pmax(indexScaled, limits[1]),limits[2]);
		p <- logodds2p(predict(bbf$loess,indexScaled));
		return(data.frame(index=index,pRough=pRough,indexScaled=indexScaled,p=p));
	} else {
		return(data.frame(index=index,p=p));
	}
}

CompareBBFGroups <- function(list,xlim=NULL,ylim=NULL,
				labels=names(list),
				col=1:length(list),pch=1:length(list),lty=1:length(list),lwd=1+0*col,
				ltyH=1:length(list),
				hbar=TRUE,
				xlab="Median of bin", ylab="Log odds",legend=TRUE,
				...) {
	if (is.null(xlim)) {
		xlim <- c( min(sapply(list, function(x) { min(x$med) ;})),
				max(sapply(list, function(x) { max(x$med) ;})) );
	}
	if (is.null(ylim)) {
		ylim <- c( min(sapply(list, function(x) { min(logodds(x$p)); })),
				max(sapply(list, function(x) { max(logodds(x$p)); })) );
	}
	for (i in 1:length(names(list))) {
		name <- names(list)[i];
		y <- logodds(list[[name]]$p);
		if (i == 1) {
			plot(list[[name]]$med,y,col=col[i],pch=pch[i],lty=lty[i],lwd=lwd[i],
				xlim=xlim,ylim=ylim,type="b",xlab=xlab,ylab=ylab,...);
		} else {
			points(list[[name]]$med,y,col=col[i],pch=pch[i],lty=lty[i],lwd=lwd[i],type="b");
		}
		if (hbar) {
			linePairs(list[[name]]$min, y, list[[name]]$max, y, col=col[i],lwd=lwd[i],lty=ltyH[i]);
		}
	}
	if(legend) {
		legend(mean(xlim),max(ylim),yjust=1,labels,pch=pch,col=col,lty=lty,lwd=lwd);
	}
}

# like BinnedBinaryFit but where x is a factor. Returns a data frame of p, nT, nF, levels
ClassFit <- function(x,y,fT=NULL,pActual=.5,a=0,b=1,c=1,d=0,
			alpha=2,beta=2,plot=FALSE,debug=FALSE,
			xlab="",main="",log="") {
	if (!is.null(fT)) {
		# notice this is a no-op if fT==pActual
		ratio <- (pActual/(1-pActual)) * (1-fT)/fT;
		a = 0; b <- 1; c <- ratio; d <- 1-c;
	}
	nT <- nF <- p <- rep(0, length(levels(x)));
	for (i in 1:length(levels(x))) {
		nT[i] <- countif(codes(x)==i & y);
		nF[i] <- countif(codes(x)==i & !y);
		p[i] <- pseudocountLL(nT[i],nF[i],a,b,c,d,alpha=alpha,beta=beta,debug=debug)$p;
	}
	if (plot) {
		plot(1:length(levels(x)), p,main=main,log=log,xlab=xlab);
	}
	return(data.frame(p=p,nT=nT,nF=nF,levels=levels(x)));
}

BayesianPredict <- function(class,value,vars,data,fT=countPerIf(data[[class]]==value)) {
	bdata <- list();
	bdata[[class]] <- ifelse(data[[class]]==value,1,0);
	formula <- paste(class, sep="");
	for (v in vars) {
		p <- BinnedBinaryFit(data[[v]], data[[class]]==value, fT=fT)$p;
		bdata[[v]] <- log(p/(1-p));
		formula <- paste(formula, ifelse(v==vars[1], "~", "+"),v,sep=" ");
	}
	bdata <- data.frame(bdata);
	formula <- eval(parse(text=formula));
	glm <- glm(formula,family=binomial(logit),data=bdata);
	return(list( glm=glm, data=bdata, cor=cor(fitted(glm),bdata[[class]]) ));
}

GenomeMedianSepAll <- function(gncDir="operons/out",prefix="gnc",genomes=NULL,...) {
	pairs <- read.delim(paste(gncDir,paste(prefix,"AllGenomes",sep=""),sep="/"));
	perm <- order(pairs$TaxId1);
	if (is.null(genomes)) genomes <- unique(sort(pairs$TaxId1));
	names <- pairs$Genome1[perm[findInterval(genomes,sort(pairs$TaxId1))]];
	out <- 0*(1:length(genomes));
	for (i in 1:length(genomes)) {
		gcN <- read.delim(paste(gncDir,"/",prefix,genomes[i],".cai",sep=""));
		out[i] <- median(select(gcN$Sep,gcN$Code=="Convergent"));
	}
	return(out);
}

OperonsPredictAll <- function(gncDir="operons/out",prefix="gnc",plot="sep",save=FALSE,genomes=NULL,...) {
	pairs <- read.delim(paste(gncDir,paste(prefix,"AllGenomes",sep=""),sep="/"));
	perm <- order(pairs$TaxId1);
	if (is.null(genomes)) genomes <- unique(sort(pairs$TaxId1));
	names <- pairs$Genome1[perm[findInterval(genomes,sort(pairs$TaxId1))]];

	if (!is.null(plot) && plot != FALSE) {
		old.par <- par(no.readonly=TRUE);
		on.exit(par(old.par));
		par(mfrow=c(12,12),mar=c(1,1,1,1)+.15,ps=8);
	}
	out <- NULL;
	for (i in 1:length(genomes)) {
		cat("Running genome ",genomes[i]," ",levels(names)[names[i]],"\n");
		row <- OperonsPredict(genomes[i],levels(names)[names[i]],
			gncDir=gncDir,plot=plot,plotDetails=FALSE,...);
		if (!is.null(row)) {
			if(save) {
				writeDelim(row$gcSC,paste(gncDir,"/",prefix,row$genome,".gc",sep=""));
				writeDelim(row$preds,paste(gncDir,"/",prefix,row$genome,".pred",sep=""));
				writeDelim(row$bbfD$groups,paste(gncDir,"/",prefix,row$genome,".sep",sep=""));
				if (!is.null(row$bbfCAI$groups)) {
					writeDelim(row$bbfCAI$groups,paste(gncDir,"/",prefix,row$genome,".caidxdy",
							sep=""));
				}
			}
			# too complex to put in the data frame
			row <- without(row,row$without);
			if (is.null(out)) {
				row$name=factor(row$name,levels=levels(names));
				row$control=factor(row$control);
				row$formula=factor(row$formula);
				out <- data.frame(row);
			} else {
				out <- rbind(out,row);
			}
		}

	}
	if(save) { writeDelim(out,paste(gncDir,"/",prefix,"AllGenomes.stats",sep="")); }
	return(out);
}

# Ten-fold cross-validation...
PredictOperonsTrained <- function(predsS,predsD,fOperon=.5,
					all=NULL, #option to make predictions for all the values
					formula=(same ~ bbfGNScore+bbfGNWithin+bbfGNMinus+cfCOG),
					vars=c("GNScore","GNWithin","GNMinus","GNAll"),
					postCAIdxdy=TRUE,
					useCOGsim=TRUE,
					smooth=TRUE,
					nfold=10,seed=.Random.seed,debug=FALSE) {
	# we need to overwrite these...
	predsS$same = NULL;
	predsD$same = NULL;
	if(debug) { cat("Formula",paste(deparse(formula)),
			"vars",vars,
			"postCAIdxdy",postCAIdxdy,"seed start",seed[1],"fOp",fOperon,
			"nS", length(predsS$GNScore), "nD", length(predsD$GNScore), "\n");
			cat("Names",names(predsS),names(predsD),"\n");
	}
	save.seed <- seed;
	set.seed(seed);
	set <- order(runif(length(c(predsS$GNScore,predsD$GNScore)))) %% nfold;
	preds <- rbind(cbind(predsS,same=1),cbind(predsD,same=0));
	preds <- cbind(preds,pTrained=-1,pTrained1=-1,pTrainedD=-1,pTrainedSame=-1,pTrainedCAI=-1);
	for (var in vars) {
		preds[[paste("bbfTrained",var,sep="")]] <- 0*preds$same;
	}
	if (!is.null(all)) {
		logoddsAll <- logoddsAllD <- logoddsAllCAI <- logoddsAllLogistic <- 0*all$Gene1;
	}
	out <- NULL;
	for (i in 0:(nfold-1)) {
		train <- select(preds,set != i);
		test <- select(preds, set == i);
		pCoverS <- countif(train$same==0)/countif(train$same==1);

		# do the bbf thing
		for (var in vars) {
			if(debug) {cat("BBF",var,range(train[[var]]),"\n"); }
			bbf <- BinnedBinaryFit(train[[var]], train$same, fT=1/(1+pCoverS),
						smooth=smooth,span=1,
						plot=debug,xlab=var);
			if(debug) {cat(var,range(bbf$p),"\n");}
			bbfvar <- paste("bbf",var,sep="");
			train[[bbfvar]] <- log(bbf$p/(1-bbf$p));
			testP <- BBFPredict(test[[var]],bbf)$p;
			test[[bbfvar]] <- log(testP/(1-testP));
			out[[bbfvar]] <- bbf;
			if(countif(is.na(test[[bbfvar]])) > 0 | countif(is.na(train[[bbfvar]])) > 0) {
				cat("Bad bbf in PredictOperonsTrained\n");
				if(debug) {
					return(list(bbf=bbf,testVar=test[[var]]));
				} else {
					return(NULL);
				}
			}
			if (!is.null(all)) {
				all[[bbfvar]] <- logodds(BBFPredict(all[[var]],bbf)$p);
			}
		}
		if (useCOGsim) {
			cfCOG <- ClassFit(train$COGsimClass, train$same);
			train$cfCOG <- logodds(cfCOG$p)[codes(train$COGsimClass)];
			test$cfCOG <- logodds(cfCOG$p)[codes(test$COGsimClass)];
		} else {
			cfCOG <- NULL;
		}
		if(debug) cat("nTrain",length(preds$same),"nTrainS",countif(train$same==1),
					"nTrainD",countif(train$same==0),"\n");
		gcm <- glm(formula,family=binomial(logit),data=train);
		if(debug) { s <- summary(gcm); s$family <- NULL; s$call <- paste(deparse(s$call));
				s$terms <- paste(deparse(s$terms)); for(n in names(s)) { cat(n,s[[n]],"\n"); } }
		# now a misnomer...response is so we get p-values not log-odds
		if(debug) { cat("do predict",apply(test,2,length),"\n"); }
		pSame <- predict(gcm,test,type="response");
		if(debug) { cat("did predict",length(pSame),length(test$cfCOG),"\n"); }
		if (length(pSame) != length(test$cfCOG)) {
			if(debug) { cat("Returning bad gcm, test\n"); return(list(gcm=gcm,test=test)); }
			cat("Fatal error in Trained predictions...\n");
			return(NULL);
		}
		# fT is like fHighS before - but no mixing because we know!
		if(debug) { cat("Do bbfD\n"); }
		bbfD <- BinnedBinaryFit(train$Sep,train$same,fT=countPerIf(train$same==1),
						xout=test$Sep,smooth=smooth);
		pSameD <- bbfD$p; # also now a misnomer
		if (postCAIdxdy) {
			bbfCAI <- BinnedBinaryFit(train$CAIdxdyR-train$CAIdsqrR, train$same,
							fT=countPerIf(train$same==1),
							xout=test$CAIdxdyR-test$CAIdsqrR,smooth=smooth);
			pCAI <- bbfCAI$p;
			if(debug) { cat("LCAI",length(train$CAIdxdyR),length(train$CAIdsqrR),length(train$same),
						length(test$CAIdxdyR),length(test$CAIdsqrR),"\n");
				cat("LCAI\n"); }
		} else {
			bbfCAI <- NULL;
			pCAI <- .5 + 0*pSameD;
		}
		if(debug) { cat("L2",length(pCoverS),length(pSame),length(test$bbfGNScore),countif(is.na(pSame)),
					length(pSameD),length(pCAI),
					length(test$cfCOG), countif(set==i),"\n"); }
		pOpLogistic <- combineP(pCoverS/(1+pCoverS), pSame);
		pOp1 <- combineP(pOpLogistic, pSameD);
		pOp <- combineP(pOp1, pCAI);
		if(debug) { cat("L2",length(pOp1),length(pOpLogistic),length(pOp),"\n"); cat("L2\n"); }
		preds$pTrained[select(1:length(preds$pOp), set==i)] <- pOp;
		preds$pTrainedOpLogistic[select(1:length(preds$pOp), set==i)] <- pOpLogistic;
		preds$pTrained1[select(1:length(preds$pOp1), set==i)] <- pOp1;
		preds$pTrainedD[select(1:length(preds$pOp), set==i)] <- pSameD;
		preds$pTrainedSame[select(1:length(preds$pOp), set==i)] <- pSame;
		preds$pTrainedCAI[select(1:length(preds$pOp), set==i)] <- pCAI;
		preds$cfCOGTrained[select(1:length(preds$pOp), set==i)] <- test$cfCOG;
		for (var in vars) {
			preds[[paste("bbfTrained",var,sep="")]][select(1:length(preds$pOp), set==i)] <-
				test[[paste("bbf",var,sep="")]];
		}
		if (!is.null(all)) {
			# note that the bbfvars were modified above to contain values based on new models...
			all$cfCOG <- logodds(cfCOG$p)[codes(all$COGsimClass)];
			pAllLogistic <- combineP(pCoverS/(1+pCoverS),predict(gcm,all,type="response"));
			pAllD <- BBFPredict(all$Sep, bbfD)$p;
			pAllCAI <- if (postCAIdxdy) BBFPredict(all$CAIdxdyR-all$CAIdsqrR, bbfCAI)$p else 0;
			logoddsAll <- logoddsAll +
				(logodds(pAllLogistic) + logodds(pAllD) + logodds(pAllCAI))/nfold;
			logoddsAllLogistic <- 	logoddsAllLogistic + logodds(pAllLogistic)/nfold;
			logoddsAllD <- logoddsAllD + logodds(pAllD)/nfold;
			logoddsAllCAI <- logoddsAllCAI + logodds(pAllCAI)/nfold;
		}
		if(debug) { cat("Finished round",i,"\n");  }
	}
	if (!is.null(all)) {
		pAll <- exp(logoddsAll)/(1+exp(logoddsAll));
		pAllLogistic <- exp(logoddsAllLogistic)/(1+exp(logoddsAllLogistic));
		pAllD <- exp(logoddsAllD)/(1+exp(logoddsAllD));
		pAllCAI <- exp(logoddsAllCAI)/(1+exp(logoddsAllCAI));
		index <- findMatch(preds$Gene1, all$Gene1)$index;
		pAll[select(index,index>0)] <- select(preds$pTrained,index>0);
		pAllLogistic[select(index,index>0)] <- select(preds$pTrainedOpLogistic,index>0);
		pAllD[select(index,index>0)] <- select(preds$pTrainedD,index>0);
		pAllCAI[select(index,index>0)] <- select(preds$pTrainedCAI,index>0);
	}
	out <- c(out, list(predsS=select(preds,preds$same==1),predsD=select(preds,preds$same==0),
				# preds for all same-pair genes
				pAll=pAll, pAllLogistic=pAllLogistic, pAllD=pAllD, pAllCAI=pAllCAI,
				bbfD=bbfD, bbfCAI=bbfCAI, cfCOG=cfCOG, gcm=gcm,
				seed=save.seed, vars=vars ));
	if(debug) { cat("Finished Training\n");  cat("Done\n"); }
	return(c(out, list(without=c("without",names(out)),
			nfold=nfold, formula=paste(deparse(formula)), postCAIdxdy=postCAIdxdy, 
			corP=cor(preds$pTrained,preds$pOp),
			diffP=mean(abs((preds$pTrained-preds$pOp))),
			corLogistic=cor(preds$pTrainedOpLogistic,preds$pOpLogistic),
			diffLogistic=mean(abs((preds$pTrainedOpLogistic-preds$pOpLogistic))),
			corDist=cor(preds$pOpDistance,preds$pTrainedD),
			diffDist=mean(abs((preds$pTrainedD-preds$pOpDistance))),
			corCAI=cor(preds$pCAI,preds$pTrainedCAI),
			pCoverS=pCoverS,
			diffCAI=mean(abs((preds$pTrainedCAI-preds$pCAI)))
	)));
}

# Takes as input OperonsPredict's output and a list of Gene1's for operon and for non-operon pairs
OperonsAOC <- function(op, same=NULL, non=NULL,
			plot="all", seed=.Random.seed, debug=FALSE, nfold=10) {
	if(is.null(same) || is.null(non)) {
		if (op$preds$Taxon[1] == 83333) {
			same = read.delim("operons/data/ecNewSameOp")$Gene1;
			non = read.delim("operons/data/ecNewAdj")$Gene1;
		} else if (op$preds$Taxon[1] == 1423) {
			same = read.delim("operons/data/bsSameOp")$Gene1;
			non = read.delim("operons/data/bsAdjNotOp")$Gene1;
		}
		else {
			cat("Needs same and non!\n"); return(NULL);
		}
	}
	same <- unique(sort(same));
	non <- unique(sort(non));
	known <- 0*op$preds$pOp;
	known[select(1:length(known),findMatch(op$preds$Gene1,same)$index>0)] <- 1;
	known[select(1:length(known),findMatch(op$preds$Gene1,non)$index>0)] <- -1;

	predsS <- select(op$preds, findMatch(op$preds$Gene1,same)$index > 0); # same operon
	if (length(predsS$Gene1) != length(same)) {
		cat("Warning: not all same-operon ids matched!\n");
	}
	predsD <- select(op$preds, findMatch(op$preds$Gene1,non)$index > 0); # different operon
	if (length(predsD$Gene1) != length(non)) {
		cat("Warning: not all non-operon ids matched!\n");
	}
	runif(1); # force seed creation
	trained <- PredictOperonsTrained(predsS, predsD, fOperon=op$fOperon, seed=seed, debug=debug,
					formula=eval(parse(text=op$formula)),
					vars=op$vars,
					postCAIdxdy=op$postCAIdxdy,
					smooth=op$smooth,
					all=op$preds, nfold=nfold );
	if(debug) { cat("PredictOperonsTrained returned",range(trained$predsS$pTrained),
				range(trained$predsD$pTrained),range(predsS$pOp),range(predsS$GNScore),"\n"); }
	aocTrained <- AOC(trained$predsS$pTrained,trained$predsD$pTrained);
	aocUntrained <- AOC(predsS$pOp,predsD$pOp);
	aocUntrained1 <- AOC(predsS$pOp1,predsD$pOp1);
	if(debug) { cat("3 AOC done"); }

	# Naive is to use GNScore > 0 or smaller sep is better (up to -20)
	naive <- function(preds) {
		return(ifelse(preds$GNScore > 0, -1, ifelse(preds$Sep< -20, 1e4, preds$Sep)));
	}
	aocNaive <- AOC(naive(predsS),naive(predsD),bigger=FALSE);
	aocDistance <- AOC(predsS$pSameD,predsD$pSameD);
	aocPSame <- AOC(predsS$pSame,predsD$pSame);
	if(debug) { cat("6 AOC done\n"); }
	if (plot=="all") {
		old.par <- par(no.readonly=TRUE);
		on.exit(par(old.par));
		par(mfrow=c(2,1));
	}
	if (plot=="all" || plot=="hist") {
		CompareHistogramList(list(Operon=predsS$pOp,Adjacent=predsD$pOp),breaks=c(0:20)/20,
					xlab="p(Operon) for known-operon and adjacent pairs",
					main=op$name);
	}
	if(debug) { cat("2 plots done\n"); }
	if (plot=="all" || plot=="AOC") {
		plot(1-aocUntrained$spec,aocUntrained$sens,type="l",lty=1,col=1,xlab="FP",ylab="TP",
			main=paste(op$name,"--",op$control,"control","--",op$formula,
					if(op$postCAIdxdy) "then CAIdxdy" else "",
					"\nAOC =",format(aocUntrained$AOC,digits=3)));
		if(debug) { cat("count",countPerIf(predsD$bOp),"\n"); }
		points(countPerIf(predsD$bOp), countPerIf(predsS$bOp), col=1, pch=2);
		points(1-countPerIf(predsD$pOp<.9), countPerIf(predsS$pOp>=.9), col=1, pch=3);
		lines(1-aocNaive$spec, aocNaive$sens, lty=4, col=2);
		points(countPerIf(predsD$GNScore>=23 | (predsD$Sep>=-20 & predsD$Sep < 55)),
			countPerIf(predsS$GNScore>=23 | (predsS$Sep>=-20 & predsS$Sep < 55)),
			pch=5, col=3);
		lines(1-aocDistance$spec, aocDistance$sens, lty=6, col=4);
		points(1-countPerIf(predsD$pSameD<.5), countPerIf(predsS$pSameD>=.5), col=4,pch=20);
		lines(1-aocPSame$spec, aocPSame$sens, lty=7, col=5);
		fH <- 1/(1+op$pCoverS);
		points(1-countPerIf(predsD$pSame<fH), countPerIf(predsS$pSame>=fH), col=5,pch=20);
		lines(1-aocTrained$spec, aocTrained$sens, lty=8, col=6, lwd=2);
		points(1-countPerIf(trained$predsD$pTrained<.5), countPerIf(trained$predsS$pTrained>=.5),
			col=6,pch=20);
		vline(0);hline(1);
		legend(1,.8,c("Untrained", "Untrained p>=fOp", "Untrained p>=.9",
				"GNScore > 0 or Sep in [-20,N)",
				"GNScore >= 23 or Sep in [-20,55)",
				"Distance only (after logistic)",
				"pSame only",
				"Trained"),
			pch=c(30,2,3,30,5,30,30,30),lty=c(1,0,0,4,0,6,7,8),col=c(rep(1,3),2:6),
			lwd=c(rep(1,7),2),xjust=1);
		f <- function(x) { return(format(x,digits=3)); }
		cat("Untrained AOC\t", f(aocUntrained$AOC),
					"\tmax acc\t", f(0.5*max(aocUntrained$spec+aocUntrained$sens)),
					"\tdef sens\t", f(countPerIf(predsS$bOp)),
					"\tspec\t",f(countPerIf(!predsD$bOp)), "\n",sep="");
		cat("pSame&Dist AOC\t", f(aocUntrained1$AOC),
					"\tmax acc\t", f(0.5*max(aocUntrained1$spec+aocUntrained1$sens)),
					"\tdef sens\t", f(countPerIf(predsS$bOp1)),
					"\tspec\t",f(countPerIf(!predsD$bOp1)),"\n",sep="");
		cat("Dist-only AOC\t",f(aocDistance$AOC),
					"\tmax acc\t", f(0.5*max(aocDistance$spec+aocDistance$sens)),
					"\tdef sens\t",f(countPerIf(predsS$pSameD>=.5)),
					"\tspec\t",f(countPerIf(predsD$pSameD<.5)),"\n",sep="");
		cat("pSame-only AOC\t",f(aocPSame$AOC),
					"\tmax acc\t", f(0.5*max(aocPSame$spec+aocPSame$sens)),
					"\tdef sens\t",f(countPerIf(predsS$pSame>=fH)),
					"\tspec\t",f(countPerIf(predsD$pSame<fH)),"\n",sep="");
		cat("Trained AOC\t",f(aocTrained$AOC),
					"\tmax acc\t", f(0.5*max(aocTrained$spec+aocTrained$sens)),
					"\tdef sens\t", f(countPerIf(trained$predsS$pTrained>=.5)),
					"\tspec\t", f(countPerIf(trained$predsD$pTrained<.5)),
					"\nGNScore>0 AOC\t",f(aocNaive$AOC), "\n", sep="");
	}
	out <- list(predsS=trained$predsS, predsD=trained$predsD,
		aocUntrained=aocUntrained,aocNaive=aocNaive,aocDistance=aocDistance,aocPSame=aocPSame,
		aocTrained=aocTrained,seed=seed,nfold=trained$nfold,
		gcm=trained$gcm,
		bbfD=trained$bbfD, bbfCAI=trained$bbfCAI,cfCOG=trained$cfCOG,
		defSens=countPerIf(predsS$bOp), defSpec=countPerIf(!predsD$bOp),
		maxAcc=.5*max(aocUntrained$spec+aocUntrained$sens),
		expectedDRatio = (1 - countPerIf(!predsS$bOp) - countPerIf(predsD$bOp)),
		known=known,
		pTrained=trained$pAll, pTrainedD=trained$pAllD, pTrainedLogistic=trained$pAllLogistic,
		pTrainedCAI=trained$pAllCAI,
		trained=data.frame(without(trained,trained$without)));
	for (var in op$vars) {
		bbfvar <- paste("bbf",var,sep="");
		out[[bbfvar]] <- trained[[bbfvar]];
	}
	return(out);
}

# handles output from OperonsAOC
OperonsCompareAOC <- function(aoc1,aoc2,var="pOp",plot=FALSE) {
	return(CompareAOC(aoc1$predsS[[var]],aoc1$predsD[[var]],aoc2$predsS[[var]],aoc2$predsD[[var]],plot=plot));
}

CompareAOC2 <- function(aoc,var1="pOp",var2="pTrained",plot=FALSE) {
	return(CompareAOC(aoc$predsS[,var1],aoc$predsD[,var1],
			aoc$predsS[,var2],aoc$predsD[,var2],plot=plot));
}

AOCStandardError <- function(s1,d1,nOn=length(s1),nOff=length(d1),area=makeaoc(s1,d1)) {
	Q1 <- area/(2-area);
	Q2 <- 2*area*area/(1+area);
	top <- area*(1-area) + (nOn-1)*(Q1-area*area) + (nOff-1)*(Q2-area*area);
	return( sqrt(top/(nOn*nOff)) );
}

# Uses the method of DeLong et al 1988:
# DeLong ER, DeLong DM, Clarke-Pearson DL.
# Biometrics. 1988 Sep;44(3):837-45.
# Comparing the areas under two or more correlated receiver operating characteristic curves:
# a nonparametric approach.
CompareAOC3 <- function(m1,m2,truth,...) {
	return(CompareAOC(select(m1,truth),select(m1,!truth),select(m2,truth),select(m2,!truth),...));
}

CompareAOC <- function(s1,d1,s2,d2,plot=FALSE, col=1:2, lty=1:2, ...) {
	p1 <- findIntervalAvg(s1, sort(d1))/length(d1); # s1 in d1
	q1 <- 1 - findIntervalAvg(d1, sort(s1))/length(s1); # d1 in s1
	area1 <- mean(p1);

	p2 <- findIntervalAvg(s2, sort(d2))/length(d2); # s2 in d2
	q2 <- 1 - findIntervalAvg(d2, sort(s2))/length(s2); # d2 in s2
	area2 <- mean(p2);

	if(plot) {
		plot((1:length(p1))/length(p1), sort(p1), col=col[1], lty=lty[1], type="l", ... );
		lines((1:length(p2))/length(p2), sort(p2), col=col[2],lty=lty[2], ... );
	}

	m <- length(s1);
	n <- length(d1);

	var1 <- (m*sum(p1**2)-sum(p1)**2)/(m*m*(m-1)) + (n*sum(q1**2)-sum(q1)**2)/(n*n*(n-1)) +
			area1*(1-area1)/(m*n);
	var2 <- (m*sum(p2**2)-sum(p2)**2)/(m*m*(m-1)) + (n*sum(q2**2)-sum(q2)**2)/(n*n*(n-1)) +
			area2*(1-area2)/(m*n);
	cov12 <- sum((p1-mean(p1))*(p2-mean(p2)))/(m*(m-1)) + sum((q1-mean(q1))*(q2-mean(q2)))/(n*(n-1));
	se <- sqrt(var1+var2-2*cov12);
	z <- (area1-area2)/se;
	return(data.frame(area1=area1,area2=area2,
			m=m,n=n,var1=var1,var2=var2,cov12=cov12,se=se,
			varp1=var(p1),varq1=var(q1),
			z=z,p=2*pnorm(abs(z),lower=FALSE)));
}

OperonUarrayCor <- function(op, orfData, aoc=NULL, skip=3, minN=10, plot=TRUE, name="orfId",spearman=FALSE) {
	i1 <- findMatch(op$preds$Gene1,orfData[,name])$index;
	i2 <- findMatch(op$preds$Gene2,orfData[,name])$index;
	preds <- select(cbind(op$preds,i1=i1,i2=i2),i1>0&i2>0);
	highOp <- rank(preds$pOpDistance) > .8*length(preds$pOpDistance) &
			rank(preds$pOpLogistic) > .8*length(preds$pOpLogistic);
	lowOp <- rank(preds$pOpDistance) < .2*length(preds$pOpDistance) &
			rank(preds$pOpLogistic) < .2*length(preds$pOpLogistic);
	cat("kept",length(preds$i1),"high",countif(highOp),"low",countif(lowOp),"\n")
	outNames <- names(orfData)[(skip+1):length(names(orfData))];
	out <- data.frame(name=outNames);
	if (!is.null(aoc)) {
		iS1a <- findMatch(aoc$predsS$Gene1,orfData$orfId)$index;
		iS2a <- findMatch(aoc$predsS$Gene2,orfData$orfId)$index;
		iS1 <- select(iS1a, iS1a>0 & iS2a>0);
		iS2 <- select(iS2a, iS1a>0 & iS2a>0);

		iD1a <- findMatch(aoc$predsD$Gene1,orfData$orfId)$index;
		iD2a <- findMatch(aoc$predsD$Gene2,orfData$orfId)$index;
		iD1 <- select(iD1a, iD1a>0 & iD2a>0);
		iD2 <- select(iD2a, iD1a>0 & iD2a>0);

		cat("knownOp",length(iS1),"adjKnownOp",length(iD1),"\n");
	}
	for (j in (skip+1):length(names(orfData))) {
		x <- orfData[,names(orfData)[j]];
		if (!is.null(aoc)) {
			out$knownOp[j-skip] = corNoNa(x[iS1],x[iS2],spearman=spearman);
			out$adjKnownOp[j-skip] = corNoNa(x[iD1],x[iD2],spearman=spearman);
		}
		x1 <- x[preds$i1];
		x2 <- x[preds$i2];
		if (countif(!is.na(x1) & !is.na(x2) & highOp) >= minN
			&& countif(!is.na(x1) & !is.na(x2) & lowOp) >= minN) {
			out$highOp[j-skip]=corNoNaSelect(x1,x2,highOp,spearman=spearman);
			out$lowOp[j-skip]=corNoNaSelect(x1,x2,lowOp,spearman=spearman);
			out$bOp[j-skip]=corNoNaSelect(x1,x2,preds$bOp,spearman=spearman);
			out$notOp[j-skip]=corNoNaSelect(x1,x2,!preds$bOp,spearman=spearman);
			out$all[j-skip]=corNoNa(x1,x2,spearman=spearman);
		}
	}
	return(out);
}

OperonUarrayAccuracy <- function(op, orfData, dims, plot=TRUE, fixedR=FALSE,name="orfId") {
	acc <- 0*dims;
	i1 <- findMatch(op$preds$Gene1, orfData[,name])$index;
	i2 <- findMatch(op$preds$Gene2, orfData[,name])$index;
	preds <- select(cbind(op$preds,i1=i1,i2=i2),i1>0&i2>0);
	highOp <- rank(preds$pOpDistance) > .8*length(preds$pOpDistance) &
			rank(preds$pOpLogistic) > .8*length(preds$pOpLogistic);
	lowOp <- rank(preds$pOpDistance) < .2*length(preds$pOpDistance) &
			rank(preds$pOpLogistic) < .2*length(preds$pOpLogistic);
	out <- data.frame(name=names(orfData)[dims],dim=dims,corLow=0*dims,corHigh=0*dims,nH=0*dims,nL=0*dims,
				accOp=0*dims,accNotOp=0*dims);
	if(plot) {
		old.par <- par(no.readonly=TRUE);
		on.exit(par(old.par));
		n <- ceiling(sqrt(1*length(dims)));
		par(mfrow=c(n,n));
	}
	for (i in 1:length(dims)) {
		x <- orfData[,dims[i]];

		xH <- select(x[preds$i1],!is.na(x[preds$i1]) & !is.na(x[preds$i2]) & highOp);
		yH <- select(x[preds$i2],!is.na(x[preds$i1]) & !is.na(x[preds$i2]) & highOp);
		xL <- select(x[preds$i1],!is.na(x[preds$i1]) & !is.na(x[preds$i2]) & lowOp);
		yL <- select(x[preds$i2],!is.na(x[preds$i1]) & !is.na(x[preds$i2]) & lowOp);

		if (length(xH) < 25 || length(yH) < 25) {
			cat("Ignoring dim",i," -- length <25\n");
			out$nH[i] <- out$nL[i] <- out$r1[i] <- out$r2[i] <- out$accOp[i] <- out$accNotOp[i] <-
					out$accOp2[i] <- out$accNotOp2[i] <- out$acc3[i] <- out$acc3D[i] <-
					out$accR3[i] <-
					out$acc3Max[i] <- out$ac3D[i] <- out$accR4[i] <- out$acc4[i] <-
					out$p.wilcox[i] <- out$p.wilcoxHL[i] <- -1;
		} else {
			out$corHigh[i] <- cor(xH,yH);
			out$nH[i] <- length(xH);
	
			out$corLow[i] <- cor(xL,yL);
			out$nL[i] <- length(xL);

			xOp <- select(x[preds$i1],!is.na(x[preds$i1]) & !is.na(x[preds$i2]) & preds$bOp);
			yOp <- select(x[preds$i2],!is.na(x[preds$i1]) & !is.na(x[preds$i2]) & preds$bOp);

			xNotOp <- select(x[preds$i1],!is.na(x[preds$i1]) & !is.na(x[preds$i2]) & !preds$bOp);
			yNotOp <- select(x[preds$i2],!is.na(x[preds$i1]) & !is.na(x[preds$i2]) & !preds$bOp);

			if (fixedR) {
				r1 <- .5; r2 <- 0;
			} else {
				r1 <- cor(xH,yH); r2 <- cor(xL,yL);
			}
			out$r1[i] <- r1;
			out$r2[i] <- r2;
			prH <- logpRatioUnnorm(xH,yH,r1,r2);
			prL <- logpRatioUnnorm(xL,yL,r1,r2);
			prOp <- logpRatioUnnorm(xOp,yOp,r1,r2);
			prNotOp <- logpRatioUnnorm(xNotOp,yNotOp,r1,r2);
			h <- CompareHistogramList(list(highOp=prH,lowOp=prL,bOp=prOp,notOp=prNotOp),
							breaks=c(-1,-.5,-.25,-.15,-.05,0,.05,.15,.25,.5,1,1.5,2,2.5),
							ylim=c(0,.3),
							plot=FALSE,returnP=TRUE,main=names(orfData)[dims[i]],
							warn=FALSE);
			model1 <- lm(bOp~lowOp+highOp+0,h,weights=lowOp+highOp+1/(length(prL)+length(prH)));
			model2 <- lm(notOp~lowOp+highOp+0,h,weights=lowOp+highOp+1/(length(prL)+length(prH)));

			out$accOp[i] <- model1$coefficients[2]/sum(model1$coefficients);
			out$accNotOp[i] <- model2$coefficients[1]/sum(model2$coefficients);

			ksR <- ksRanks(list(prH=prH,prL=prL,prOp=prOp,prNotOp=prNotOp));
			ks1 <- lm(prOp~prL+prH+0,ksR);
			ks2 <- lm(prNotOp~prL+prH+0,ksR);
			out$accOp2[i] <- ks1$coefficients[2]/sum(ks1$coefficients);
			out$accNotOp2[i] <- ks2$coefficients[1]/sum(ks2$coefficients);

			ksR <- ksRanks(list(prH=prH,prL=prL,prOp=prOp,prNotOp=prNotOp));

			ks3 <- lm(I(prOp-prNotOp) ~ I(prH-prL)+0, ksR);
			out$acc3[i] <- ks3$coefficients[1];
			out$accR3[i] <- cor(ksR$prOp-ksR$prNotOp, ksR$prH-ksR$prL);
			out$acc3Max[i] <- min(ksR$prOp-ksR$prNotOp)/min(ksR$prH-ksR$prL); # both are negative
			out$acc3D[i] <- ks.test(prOp,prNotOp)$statistic/ks.test(prH,prL)$statistic;
			if (median(prOp) < median(prNotOp) || median(prH) < median(prL) || r1 <= r2) out$acc3D[i] <- 0;

			if (FALSE) {
				plot(ksR$per,ksR$prH,type="l",col=3);
				eqline(col=1);
				lines(ksR$per,ksR$prL,type="l",col=2);
				lines(ksR$per,ksR$prOp,type="l",col=3,lty=2);
				lines(ksR$per,ksR$prNotOp,type="l",col=2,lty=2);
				lines(ksR$per, ksR$prH*out$accOp2[i] + ksR$prL*(1-out$accOp2[i]),col="grey",lty=3);
				lines(ksR$per, ksR$prL*out$accNotOp2[i] + ksR$prH*(1-out$accOp2[i]),col="grey",lty=3);
			}
			if (plot) {
				plot(ksR$per, ksR$prH-ksR$prL, type="l");
				hline(0);
				lines(ksR$per, ksR$prOp-ksR$prNotOp, col=2);
				lines(ksR$per, out$acc3[i]*(ksR$prH-ksR$prL), col="grey", lty=2);
			}
			xBoth <- select(x[preds$i1], !is.na(x[preds$i1]) & !is.na(x[preds$i2]));
			yBoth <- select(x[preds$i2], !is.na(x[preds$i1]) & !is.na(x[preds$i2]));
			pOp <- select(preds$pOp, !is.na(x[preds$i1]) & !is.na(x[preds$i2]));
			prBoth <- logpRatioUnnorm(xBoth,yBoth,r1,r2);
			out$accR4[i] <- cor(rank(pOp),rank(prBoth));
			out$acc4[i] <- lm(I(rank(prBoth)-length(prBoth)/2)~I(rank(pOp)-length(pOp)/2)+0)$coefficients[1];
			if (FALSE) {
				nBreaks <- 20;
				breaks <- (0:nBreaks) * (length(pOp) / nBreaks);
				mids <- (breaks[1:nBreaks]+breaks[2:(nBreaks+1)])/2;
				x <- aggregate(rank(prBoth),list(pOp=cut(rank(pOp),breaks=breaks)),mean);
				plot(mids,x$x,type="l");
				lines(0:length(pOp),length(pOp)/2 + ((0:length(pOp)) - length(pOp)/2)*out$acc4[i],col="grey");
			}
			out$p.wilcox[i] <- wilcox.test(prOp,prNotOp)$p.value;
			out$p.wilcoxHL[i] <- wilcox.test(prH,prL)$p.value;
		}
	}
	return(out);
}

UarrayToSums <- function(gene, orfData, dims, name="orfId", orfs=orfData[,name]) {
	i <- findMatch(gene, orfs)$index;
	out <- rep(0, length(gene));
	out[select(1:length(gene),i>0)] <- apply(orfData[select(i,i>0),dims], 1, sum, na.rm=TRUE);
	return(out);
}

UarrayOrfPairCor <- function(gene1, gene2, orfData, dims, minN=10, name="orfId", orfs=orfData[,name],debug=debug) {
	i1 <- findMatch(gene1, orfs)$index;
	i2 <- findMatch(gene2, orfs)$index;
	out <- vector(length=length(gene1));
	n <- sumD <- sumD1 <- sumDsqr <- diff2 <- diff1 <- 0*(1:length(out));

	for (i in 1:length(out)) {
		if(debug) { cat(i,"\n"); }
		if (i1[i] > 0 && i2[i] > 0) {
			dimsUse <- dims[!is.na(orfData[i1[i],dims]) & !is.na(orfData[i2[i],dims])];
			n[i] <- length(dimsUse);
			if (n[i] >= minN) {
				d1 <- t(orfData[i1[i],dimsUse]);
				d2 <- t(orfData[i2[i],dimsUse]);
				out[i] <- cor(d1,d2);
				sumD[i] <- sum(abs(d1))+sum(abs(d2));
				sumD1[i] <- sum(abs(d1));
				diff1[i] <- sum(abs(d1)-abs(d2));
				diff2[i] <- sum((d1-d2)**2);
				sumDsqr[i] <- sum(d1**2)+sum(d2**2);
			}
		} else {
			out[i] <- NA;
		}
	}
	return(data.frame(orfPairs=out,n=n,sumD=sumD,sumD1=sumD1,sumDsqr=sumDsqr,diff1=diff1,diff2=diff2,
				dist=sqrt(diff2/sumDsqr),dist1=diff1/sumD));
}

UarrayOrfPairCorAll <- function(preds, data, dims, minN=10, name="orfId",debug=debug) {
	means <- apply(data[,dims],2,function(x) { return(mean(withoutNA(x))); });
	swept <- sweep(data[,dims],2,means);
	return(UarrayOrfPairCor(preds$Gene1,preds$Gene2,swept,1:length(swept),minN=minN,orfs=data[,name],debug=debug));
}

ReadUarrayOpCors <- function(prefix,op,dir="operons/data",buildOrfPairs=TRUE,minN=10) {
	if (buildOrfPairs) {
		ratios <- read.delim(paste(dir,"/",prefix,"Ratios",sep=""));
		controlDyes <- read.delim(paste(dir,"/",prefix,"ControlDyes",sep=""));
		treatmentDyes <- read.delim(paste(dir,"/",prefix,"TreatmentDyes",sep=""));
		ses <- read.delim(paste(dir,"/",prefix,"SEs",sep=""));

		dims <- select(2:length(ratios),ses$condition != "Genomic");

		S1c <- UarrayToSums(op$gc$Gene1, controlDyes, dims);
		S1t <- UarrayToSums(op$gc$Gene1, treatmentDyes, dims);
		S2c <- UarrayToSums(op$gc$Gene2, controlDyes, dims);
		S2t <- UarrayToSums(op$gc$Gene2, treatmentDyes, dims);
		S4 <- pmax(0,pmin(S1c+S1t,S2c+S2t));
		cors <- cbind(UarrayOrfPairCorAll(op$gc, ratios, dims, minN=minN),
				S4=S4,S1c=S1c,S1t=S1t,S2c=S2c,S2t=S2t);
		cors$orfPairs[select(1:length(cors$n),cors$n < minN)] <- NA;
		writeDelim(cors,file=paste(dir,"/",prefix,"Cors",sep=""));
		return(cors);
	} else {
		cors <- read.delim(paste(dir,"/",prefix,"Cors",sep=""));
		cors$orfPairs[select(1:length(cors$n),cors$n < minN)] <- NA;
		return(cors);
	}
}

SimulateMixtureEstimation <- function(n=c(1000,100,250,250), mean2=.5, sd2=1, from=-2, to=2, tries=100) {
	out <- rep(0,tries);
	for (i in 1:tries) {
		trueOff <- pmax(from,pmin(to, rnorm(n[1]) ));
		trueOn <- pmax(from,pmin(to, rnorm(n[2],mean=mean2,sd=sd2) ));
		test <- pmax(from,pmin(to, c(rnorm(n[3]),rnorm(n[4],mean=mean2,sd=sd2)) ));
		d1 <- density(trueOff,from=from,to=to); d1$y <- d1$y/sum(d1$y);
		d2 <- density(trueOn,from=from,to=to);  d2$y <- d2$y/sum(d2$y);
		d3 <- density(test,from=from,to=to);    d3$y <- d3$y/sum(d3$y);
		out[i] <- lm(I(d3$y-d1$y)~I(d2$y-d1$y)+0)$coefficients[1];
	}
	return(out);
}

AccResampleGenes <- function(op,cors,n=NULL,useDensity=NULL,tries=100) {
	# four classes: 1 for opposing-strand, 2 for pred. not op, 3 for pred op (<.95), 4 for high confidence
	classSame <- cut(op$preds$pOp,c(0,.5,.95,1),labels=FALSE);
	class <- rep(1,nrow(cors));
	class[op$gcSC$same] <- 1+classSame;
	counts <- c(table(class));

	out <- data.frame(fOp=rep(0,tries),accT=rep(0,tries),accF=rep(0,tries));
	for (i in 1:tries) {
		# resample within each class
		index <- rep(0,nrow(cors));
		for (j in 1:4) {
			transfer <- (1:nrow(cors))[class==j];
			index[class==j] <- transfer[ 1+floor(runif(counts[j])*counts[j]) ];
		}			
		F <- fOperonVsUarray(op,cors[index,],n=n,useDensity=useDensity);
		out$fOp[i] <- lm(I(same-controlR)~I(highR-controlR)+0,F)$coefficients[1];
		out$accT[i] <- lm(I(op-controlO)~I(highO-controlO)+0,F)$coefficients[1];
		out$accF[i] <- lm(I(nop-highN)~I(controlN-highN)+0,F)$coefficients[1];
	}
	return(out);
}
		
UarrayOrfPairSubsampleAll <- function(prefix,op,dir="operons/data",minN=10) {
	ratios <- read.delim(paste(dir,"/",prefix,"Ratios",sep=""));
	ses <- read.delim(paste(dir,"/",prefix,"SEs",sep=""));
	out <- NULL;
	for (skip in levels(ses$condition)) {
		if (skip != "Genomic") {
			row <- UarrayOrfPairSubsample(skip, prefix, op, ratios=ratios, ses=ses);
			if (is.null(out)) { out <- row; } else { out <- rbind(out, row); }
		}
	}
	return(out);
}

UarrayOrfPairSubsample <- function(skip,prefix,op,dir="operons/data",minN=10,debug=FALSE,
		ratios=read.delim(paste(dir,"/",prefix,"Ratios",sep="")),
		ses=read.delim(paste(dir,"/",prefix,"SEs",sep=""))) {
	dims <- select(2:length(ratios),ses$condition != "Genomic" & ses$condition != skip);
	cat("Skipping",skip,"remaining arrays:",length(dims),"\n");
	minN <- min(minN,length(dims));
	cors <- UarrayOrfPairCorAll(op$gc, ratios, dims, minN=minN,debug=debug);
	cors$orfPairs[select(1:length(cors$n),cors$n < minN)] <- NA;

	F <- fOperonVsUarray(op,cors);
	fOp <- lm(I(same-controlR)~I(highR-controlR)+0,F)$coefficients[1];
	accT <- lm(I(op-controlO)~I(highO-controlO)+0,F)$coefficients[1];
	accF <- lm(I(nop-highN)~I(controlN-highN)+0,F)$coefficients[1];

	return(data.frame(skip=skip,fOp=fOp,accT=accT,accF=accF));
}

UarrayOrfPairSim <- function(preds, orfData, dim, name="orfId") {
	i1 <- findMatch(preds$Gene1, orfData[,name])$index;
	i2 <- findMatch(preds$Gene2, orfData[,name])$index;
	i12 <- select(cbind(preds,i1=i1,i2=i2), i1>0 & i2>0);
	withData <- select(i12, !is.na(orfData[i12$i1,dim]) & !is.na(orfData[i12$i2,dim]));
	return(cbind(withData,v1=orfData[withData$i1,dim],v2=orfData[withData$i2,dim],
			sim=orfData[withData$i1,dim]*orfData[withData$i2,dim]));
}

UarrayOrfPairTest <- function(preds, orfPairs, aoc=NULL) {
	highOp <- rank(preds$pOpDistance) > .8*length(preds$pOpDistance) &
			rank(preds$pOpLogistic) > .8*length(preds$pOpLogistic);
	lowOp <- rank(preds$pOpDistance) < .2*length(preds$pOpDistance) &
			rank(preds$pOpLogistic) < .2*length(preds$pOpLogistic);

	r9 <- rank(preds$pOp) > .95*length(preds$pOp);
	r1 <- rank(preds$pOp) < .05*length(preds$pOp);

	out <- list();
	out$dHighLow <- ks.test(select(orfPairs, !is.na(orfPairs) & highOp),
				select(orfPairs, !is.na(orfPairs) & lowOp ))$statistic;
	out$d20 <- ks.test(select(orfPairs, !is.na(orfPairs) & r9),
				select(orfPairs, !is.na(orfPairs) & r1 ))$statistic;
	out$dOpNot <- ks.test(select(orfPairs, !is.na(orfPairs) & preds$bOp),
				select(orfPairs, !is.na(orfPairs) & !preds$bOp ))$statistic;
	if (!is.null(aoc)) {
		iS <- findMatch(aoc$predsS$Gene1,preds$Gene1)$index;
		iD <- findMatch(aoc$predsD$Gene1,preds$Gene1)$index;
		out$dExperimental <- ks.test(select(orfPairs[iS], !is.na(orfPairs[iS])),
						select(orfPairs[iD], !is.na(orfPairs[iD])))$statistic;
		out$expectedDRatio <- aoc$expectedDRatio;

	}
	out$accuracyHL <- out$dOpNot/out$dHighLow;
	out$accuracy20 <- out$dOpNot/out$d20;
	out$accuracyExperimental <- out$dOpNot/out$dExperimental;
	return(data.frame(out));
}

# For each single microarray (given by dims), remove pairs where neither gene has changed
# (|Delta| < .3).
SingleUarrayOpTest <- function(preds, orfData, dims, aoc=NULL, plot=TRUE, name="orfId", minD=.3, minN=100) {
	if (plot && length(dims)>1) {
		n <- ceiling(sqrt(length(dims)));
		oldpar <- par(no.readonly=TRUE); on.exit(par(oldpar));
		par(mfrow=c(n,n),mar=c(1,1,1,1)+.15,ps=8);
	}
	known <- if(is.null(aoc)) 0 else aoc$known;

	means <- apply(orfData[,dims],2,function(x) { return(mean(withoutNA(x))); });
	swept <- sweep(orfData[,dims],2,means);

	i1 <- findMatch(preds$Gene1, orfData[,name])$index;
	i2 <- findMatch(preds$Gene2, orfData[,name])$index;
	i12 <- select(cbind(preds,i1=i1,i2=i2,known=known), i1>0 & i2>0);

	out <- NULL;
	for (i in 1:length(dims)) {
		dim <- dims[i];
		withData <- select(i12, !is.na(swept[i12$i1,i]) & !is.na(swept[i12$i2,i])
					& (abs(swept[i12$i1,i]) >= minD | abs(swept[i12$i2,i]) >= minD));
		if (is.null(withData) || nrow(withData) < minN) {
			cat("Skipping dimension",dim,"data",nrow(withData),"\n");
			if (plot && length(dims)>1) plot(0:0,0:0,pch=0);
		} else {
			delta <- cbind(withData,
					v1=swept[withData$i1,i],
					v2=swept[withData$i2,i],
					delta=abs(swept[withData$i1,i]-swept[withData$i2,i]) );
			if (plot) {
				CompareHistogramList(split(delta$delta,delta$bOp),breaks=c(0:10)/5,
							legendX=-100,pch=30,ylim=c(0,.5),lty=1,col=c(2,3));
			}

			ksAll <- ks.test(select(delta$delta,delta$bOp), select(delta$delta,!delta$bOp));

			knownOp <- select(delta$delta,delta$known==1);
			knownNop <- select(delta$delta,delta$known==-1);
			useK <- length(knownOp)>5 & length(knownNop)>5;
			ksD <- NULL;
			if (useK) ksD <- ks.test(select(delta$delta,delta$known==1),
							select(delta$delta,delta$known==-1));
			row <- data.frame(dim=dim,rows=nrow(withData),
					nOp=countif(delta$bOp),
					nNop=countif(!delta$bOp),
					nKnownOp=countif(delta$known==1),
					nKnownNop=countif(delta$known==-1),
					medianOp=median(select(delta$delta,delta$bOp)),
					medianNop=median(select(delta$delta,!delta$bOp)),
					medianKnown=if (useK) median(select(delta$delta,delta$known==1)) else NA,
					medianKnownNop=if (useK) median(select(delta$delta,delta$known==-1)) else NA,
					allD=ksAll$statistic, knownK=if(useK) ksD$statistic else NA,
					allR=cor(rank(delta$delta),rank(delta$pOp)),
					knownR=if(useK) cor(select(delta$delta,delta$known!=0),
								select(delta$known,delta$known!=0)) else NA);
			if(is.null(out)) { out <- row } else { out[nrow(out)+1,] <- row }
		}
	}
	return(out);
}

simulateOperonStrand <- function(fPnop=.7, fOpP=.6, fOpM=.5, n=1e3) {
	opR <- runif(n);
	bOpP <- runif(length(opR)) < fPnop; # used if not an operon
	strand <- op <- opR > 0; # vectors of desired length
	# the model is, we add a new gene, either to an operon or not
	# op[1] is true if 1->2 is an operon
	# same[1] is true if 1 and 2 are on the same strand
	for (i in 2:length(strand)) {
		op[i-1] <- if(strand[i-1]) opR[i] < fOpP else opR[i] < fOpM;
		strand[i] <- if (op[i-1]) strand[i-1] else bOpP[i];
	}
	strandNext <- strand[c(2:length(strand),1)];
	same <- strand == strandNext;
	return(data.frame(same=same,op=op,strand=strand,strandNext=strandNext));
}

summarizeSOPS <- function(sops) {
	return(data.frame(
		FPNO = countPerIf(select(sops$strand,!sops$op)),
		OP = countPerIf(select(sops$op,sops$strand)),
		OM = countPerIf(select(sops$op,!sops$strand)),
		FPlus = countPerIf(sops$strand),
		FPP = countPerIf(select(sops$same,sops$strand)),
		FMM = countPerIf(select(sops$same,!sops$strand)),
		OPP = countPerIf(select(sops$op,sops$same&sops$strand)),
		OMM = countPerIf(select(sops$op,sops$same&!sops$strand)),
		fOperon = OperonFreqStrandBias( countPerIf(select(sops$same,sops$strand)),
						countPerIf(select(sops$same,!sops$strand)) )
		));
}

# given P(+2|+1) and P(-2|-1), returns P(O|+1,+2)== P(O|-1,-2)
# under the assumption that they are equal...
OperonFreqStrandBias <- function(pSamePlus,pSameMinus) {
	a <- pSamePlus/pSameMinus;
	b <- -2*pSamePlus;
	c <- pSamePlus+pSameMinus-1;
	pOgivenMinus <- ifelse(b**2 - 4*a*c < 0, 0, -b - sqrt(b**2 - 4*a*c))/(2*a);
	return( pOgivenMinus/pSameMinus );
}


# Given features v1, v2, and known correlations with the dependent variable
# r1true, r2true, and conditional correlation cond12 (given true or false),
# simulate a data set and run our fancy BBF fit on it.
# Assume half of the data set is T for simplicity...
SimulateTraining <- function(r1true=.3,r2true=.3,cond12=0,n=2000,plot=TRUE,...) {
	truth <- ((1:n)%%2)==0;
	v1r <- r1true*ifelse(truth,1,0)/sd(truth) + rnorm(sd=sqrt(1-r1true**2),n=n);
	v1 <- rank(v1r);
	pv1T <- dnorm(v1r,mean=r1true/sd(truth), sd=sqrt(1-r1true**2));
	pv1F <- dnorm(v1r,mean=0,sd=sqrt(1-r1true**2));
	p1 <- pv1T/(pv1T+pv1F);
	v2r <- r2true*ifelse(truth,1,0)/sd(truth) + cond12*v1/sd(v1) +
			rnorm(sd=sqrt(1-r2true**2-cond12**2),n=n);
	v2 <- rank(v2r);
	pv2T <- dnorm(v2r,mean=r1true/sd(truth), sd=sqrt(1-r2true**2));
	pv2F <- dnorm(v2r,mean=0,sd=sqrt(1-r2true**2));
	p2 <- pv2T/(pv2T+pv2F);

	# put perfect error estimates into bbf
	high <- (v1 > n/2); # split for bbf
	fOgH <- countPerIf(truth,high);
	fOgL <- countPerIf(truth,!high);
	fHigh <- countPerIf(high);

	a <- (1-fOgH)*fHigh;
	b <- (fOgH-(1-fOgH))*fHigh;
	c <- a + (1-fOgL)*(1-fHigh);
	d <- b + (fOgL-(1-fOgL))*(1-fHigh);

	bbf <- BinnedBinaryFit(v2, high, a=a,b=b,c=c,d=d,plot=plot,smooth=FALSE,...);
	p2bbfLogOdds <- logodds(bbf$p);
	loessR <- loess(p2bbfLogOdds~rank(v2));
	p2loess <- logodds2p(predict(loessR));
	lines(v2[order(v2)], p2loess[order(v2)]);
	return(list(bbf=bbf,gc=data.frame(truth=ifelse(truth,1,0),v1=v1,v2=v2,
					v12=rank(r1true*v1+r2true*v2),
					high=ifelse(high,1,0),p1=p1,p2=p2,p12=combineP(p1,p2),
					p2bbf=bbf$p,p12bbf=combineP(p1,bbf$p),
					p2loess=p2loess,p12loess=combineP(p1,p2loess) )));
}

SimulateTrainings <- function(tries=30,...) {
	out <- NULL;
	for(i in 1:tries) {
		d <- SimulateTraining(...);
		row <- cor(d$gc)["truth",c("v1","v2","v12","p1","p2","p12","p2bbf","p12bbf","p2loess","p12loess")];
		row <- as.data.frame(as.list(row));
		if (i==1) { out <- data.frame(row); } else { out <- rbind(out,data.frame(row)); }
	}
	return(out);

}

# adds Log2Sum, Log2Diff, and Log2Norm (of 2/1)
# Use rename to ensure that the input contains Fg1, Bg1, Fg2, Bg2, and either sector or sectorX,sectorY
UarrayNormalize <- function(data, minOverBackground=2.0, rename=NULL, plot=TRUE) {
	if (!is.null(rename)) {
		data <- data.frame(RenameColumns(data, rename));
	}
	if (is.null(data$sector)) { data$sector <- interaction(data$sectorX,data$sectorY); }
	data$Log2Sum <- data$Log2Diff <- data$Log2Norm <- rep(NA,nrow(data));

	keep <- select(cbind(data,row=1:nrow(data)),
			data$Fg1>=minOverBackground*data$Bg1 &
			data$Fg2>=minOverBackground*data$Bg2 &
			!is.na(data$Fg1) & !is.na(data$Bg1) &
			!is.na(data$Fg2) & !is.na(data$Bg2) );
	keep$D1 <- keep$Fg1-keep$Bg1;
	keep$D2 <- keep$Fg2-keep$Bg2;
	keep$Log2Sum <- log(keep$D1)+log(keep$D2);
	keep$Log2Diff <- log(keep$D2)-log(keep$D1);
	if (plot) {
		plot(keep$Log2Sum,keep$Log2Diff,pch=codes(keep$sector),col=codes(keep$sector));
	}
	ks <- split(cbind(keep,row=1:nrow(keep)),keep$sector)
	for (s in names(ks)) {
		modelS <- loess(Log2Diff~Log2Sum,ks[[s]]);
		if (plot) lines(sort(ks[[s]]$Log2Sum),predict(modelS)[order(ks[[s]]$Log2Sum)]);
		ks[[s]]$Log2Norm <- ks[[s]]$Log2Diff-predict(modelS)
	}
	for (s in names(ks)) {
		data$Log2Norm[ks[[s]]$row] <- ks[[s]]$Log2Norm;
	}
	data$Log2Sum[keep$row] <- keep$Log2Sum;
	data$Log2Diff[keep$row] <- keep$Log2Diff;
	return(data);
}

# Use sectorDiv=c(14,15) and sectorOff=c(1,1) to subtract x and y by 1 and divide by 14,15
# to compute sectorX,sectorY
UarrayNormalizeDir <- function(dir, rename, pattern=".VIMSS$", suffix=".Norm",
				sectorOff=NULL, sectorDiv=NULL, plot=TRUE,
				minOverBackground=2.0) {
	files <- list.files(dir, pattern);
	if (plot) {
		n <- ceiling(sqrt(length(files))+1e-4);
		old.par <- par(no.readonly=TRUE); on.exit(par(old.par));
		par(mfrow=c(n,n),mar=c(1,1,1,1));
	}
	for (file in files) {
		raw <- read.delim(paste(dir,"/",file,sep=""), comment="");
		if(!is.null(rename)) raw <- data.frame(RenameColumns(raw,rename));
		if (!is.null(sectorDiv)) {
			raw$sectorX <- (raw$x-sectorOff[1]) %/% sectorDiv[1];
			raw$sectorY <- (raw$y-sectorOff[2]) %/% sectorDiv[2];
		} else { # in case no position info available
			if (is.null(raw$sectorX)) raw$sectorX <- 0;
			if (is.null(raw$sectorY)) raw$sectorY <- 0;
		}
		cat("Processing ",dir,"/",file,"\n",sep="");
		norm <- UarrayNormalize(raw, rename=NULL, plot=plot, minOverBackground=minOverBackground);
		norm$Ch1 <- norm$Fg1-norm$Bg1;
		norm$Ch2 <- norm$Fg2-norm$Bg2;
		writeDelim(norm,paste(dir,"/",file,suffix,sep=""));
		cat("Wrote ",dir,"/",file,suffix,"\n",sep="");
	}
}

ShowCorDensities <- function(op,cors,name="orfPairs",aoc=NULL,from=-1,to=1,acc=.9,ylim=c(0,3)) {
	both <- cbind(op$preds,select(cors,op$gcSC$same));
	if(!is.null(aoc)) { both$known <- aoc$known; } else { both$known <- 0; }
	data <- list(predOp=select(both,both$bOp & !is.na(both$orfPairs)),
			predNot=select(both,!both$bOp & !is.na(both$orfPairs)),
			control=select(cbind(op$gcSC,cors),!op$gcSC$same & !is.na(cors$orfPairs)));
	if (!is.null(aoc)) {
		data$knownOp <- select(both,both$known==1 & !is.na(both$orfPairs));
		data$knownNot <- select(both,both$known==-1 & !is.na(both$orfPairs));
	}
	CompareDensities(lapply(data,function(x){x[[name]]}),xlim=c(from,to),ylim=ylim);
	d <- lapply(data,function(x){density(x[[name]],from=from,to=to)});
	if (!is.null(aoc)) {
		acc2 <- 1-acc;
		lines(d$predOp$x,acc*d$knownOp$y+acc2*d$knownNot$y,col="grey",lty=2,lwd=2);
		lines(d$predOp$x,acc2*d$predOp$y+acc*d$control$y,col="grey",lty=2,lwd=2);
	}
	return(data);
}

fOperonVsUarray <- function(op,cors,fOperon=op$fOperon,n=NULL,useDensity=NULL) {
	if(is.null(n)) n <- if(is.null(op$sumD1)) 1 else 4;
	if(is.null(useDensity)) useDensity <- TRUE;

	P2 <- cbind(op$preds,select(cors,op$gcSC$same));
	P2 <- P2[!is.na(P2$orfPairs),];
	control <- select(cors,!op$gcSC$same & !is.na(cors$orfPairs));
	high <- select(P2,P2$pOp>.95);
	breaks <- quantile(select(cors$sumD1,!is.na(cors$orfPairs)),probs=seq(0,1,1/n));

	if (max(breaks)-min(breaks) > 0) {
		cutH <- cut(high$sumD1,breaks,labels=FALSE);
		cutC <- cut(control$sumD1,breaks,labels=FALSE);
		cutP2 <- cut(P2$sumD1,breaks,labels=FALSE);
		countsAll <- table(cut(P2$sumD1,breaks,labels=FALSE));
		countsOp <- table(cut(select(P2$sumD1,P2$bOp),breaks,labels=FALSE));
		countsNop <- table(cut(select(P2$sumD1,!P2$bOp),breaks,labels=FALSE));
	} else {
		cutH <- rep(1,nrow(high));
		cutC <- rep(1,nrow(control));
		cutP2 <- rep(1,nrow(P2));
		countsAll <- nrow(P2);
		countsOp <- countif(P2$bOp);
		countsNop <- countif(!P2$bOp);
	}

	histFunc <- function(x,from=-1,to=1,by=.1) {
		h <- hist(x,breaks=seq(from,to,by),plot=FALSE);
		return(list(x=h$mids,y=h$counts));
	}

	densFunc <- if(useDensity) density else histFunc;
	highModel <- lapply(split(high$orfPairs,cutH),densFunc,from=-1,to=1);
	controlModel <- lapply(split(control$orfPairs,cutC),densFunc,from=-1,to=1);
	sameModel <- densFunc(P2$orfPairs,from=-1,to=1);

	sameModel$y <- sameModel$y/sum(sameModel$y);
	opModel <- densFunc(select(P2$orfPairs,P2$bOp),from=-1,to=1);
	opModel$y <- opModel$y/sum(opModel$y);
	nopModel <- densFunc(select(P2$orfPairs,!P2$bOp),from=-1,to=1);
	nopModel$y <- nopModel$y/sum(nopModel$y);

	x <- highModel[[1]]$x;
	allR <- highR <- controlR <- highO <- controlO <- highN <- controlN <- 0*highModel[[1]]$x;
	for (i in 1:length(highModel)) {
		highModel[[i]]$y <- highModel[[i]]$y/sum(highModel[[i]]$y);
		controlModel[[i]]$y <- controlModel[[i]]$y/sum(controlModel[[i]]$y);

		allR <- allR + (countsAll[i]/sum(countsAll))*
			(fOperon*highModel[[i]]$y + (1-fOperon)*controlModel[[i]]$y);
		highR <- highR + (countsAll[i]/sum(countsAll))*highModel[[i]]$y;
		controlR <- controlR + (countsAll[i]/sum(countsAll))*controlModel[[i]]$y;

		highO <- highO + (countsOp[i]/sum(countsOp))*highModel[[i]]$y;
		controlO <- controlO + (countsOp[i]/sum(countsOp))*controlModel[[i]]$y;

		highN <- highN + (countsNop[i]/sum(countsNop))*highModel[[i]]$y;
		controlN <- controlN + (countsNop[i]/sum(countsNop))*controlModel[[i]]$y;
	}
	data <- list(x=x,mix=allR,same=sameModel$y,highR=highR,controlR=controlR,
			op=opModel$y, nop=nopModel$y,
			highO=highO, controlO=controlO,
			highN=highN, controlN=controlN,
			high1=highModel[[1]]$y,control1=controlModel[[1]]$y);
	if (max(breaks)-min(breaks)>0 && n == 4) {
		data$high2 <- highModel[[2]]$y;
		data$control2 <- controlModel[[2]]$y;
		data$high3 <- highModel[[3]]$y;
		data$control3 <- controlModel[[3]]$y;
    		data$high4 <- highModel[[4]]$y;
		data$control4 <- controlModel[[4]]$y;
	}
	return(data.frame(data));
}
